/**
*  @filename    sdk.js
*  @author      jaenster, theBGuy
*  @desc        development tool for readability
*  @format      sdk.objprop.objprop.ObjProp (excludes functions which use sdk.objprop.camelCase)
*  @type        UMD module
*
*/

(function (root, factory) {
  if (typeof module === "object" && typeof module.exports === "object") {
    let v = factory();
    if (v !== undefined) module.exports = v;
  } else if (typeof define === "function" && define.amd) {
    define([], factory);
  } else {
    root.sdk = factory();
  }
}(this, function () {
  "use strict";
  /**
   * @exports sdk
   */
  const sdk = {
    waypoints: {
      Ids: [119, 145, 156, 157, 237, 238, 288, 323, 324, 398, 402, 429, 494, 496, 511, 539],
      Act1: [0x01, 0x03, 0x04, 0x05, 0x06, 0x1b, 0x1d, 0x20, 0x23],
      Act2: [0x28, 0x30, 0x2a, 0x39, 0x2b, 0x2c, 0x34, 0x4a, 0x2e],
      Act3: [0x4b, 0x4c, 0x4d, 0x4e, 0x4f, 0x50, 0x51, 0x53, 0x65],
      Act4: [0x67, 0x6a, 0x6b],
      Act5: [0x6d, 0x6f, 0x70, 0x71, 0x73, 0x7b, 0x75, 0x76, 0x81]
    },

    difficulty: {
      Normal: 0,
      Nightmare: 1,
      Hell: 2,
      Difficulties: ["Normal", "Nightmare", "Hell"],

      nameOf: function (difficulty) {
        if (difficulty === undefined || typeof difficulty !== "number") return false;
        if (difficulty < 0 || difficulty > 2) return false;
        return ["Normal", "Nightmare", "Hell"][difficulty];
      }
    },

    party: {
      NoParty: 65535,
      flag: {
        Invite: 0,
        InParty: 1,
        Accept: 2,
        Cancel: 4
      },
      controls: {
        Hostile: 1,
        InviteOrCancel: 2,
        Leave: 3,
        Ignore: 4,
        Squelch: 5,
      },
    },

    clicktypes: {
      click: {
        item: {
          Left: 0,
          Right: 1,
          ShiftLeft: 2, // For belt
          MercFromBelt: 3, // For belt
          Mercenary: 4 // Give to merc
        },
        map: {
          LeftDown: 0,
          LeftHold: 1,
          LeftUp: 2,
          RightDown: 3,
          RightHold: 4,
          RightUp: 5,
        },
      },
      shift: {
        NoShift: 0,
        Shift: 1
      }
    },
    /**
    *  @notes
    *   - I get cursortype 3 when I swap an item that is equipped, but I get 4 if I just unequip an item
    *   - I get cursortype 3 if I pick an item from my inventory then hover it over another item
    *   - I get cursortype 4 if I pick and item from my inventory and don't hover it over another item
    *     - If I then hover it over an item it turns to 3 then stays 3
    */
    cursortype: {
      Empty: 1,
      ItemOnUnitHover: 3, // see notes
      ItemOnCursor: 4, // see notes
      Identify: 6,
      Repair: 7,
    },

    collision: {
      BlockWall: 0x01,
      LineOfSight: 0x02,
      Ranged: 0x04,
      PlayerToWalk: 0x08,
      DarkArea: 0x10,
      Casting: 0x20,
      Unknown: 0x40,
      Players: 0x80,
      Monsters: 0x100,
      Items: 0x200,
      Objects: 0x400,
      ClosedDoor: 0x800,
      IsOnFloor: 0x1000,
      MonsterIsOnFloor: 0x1100,
      MonsterIsOnFloorDarkArea: 0x1110, // in doorway
      FriendlyNPC: 0x2000,
      Unknown2: 0x4000,
      DeadBodies: 0x8000,
      MonsterObject: 0xFFFF,
      BlockMissile: 0x80E,
      WallOrRanged: 0x5,
      BlockWalk: 0x1805,
      FriendlyRanged: 0x2004,
      BoneWall: 4352,
    },

    areas: {
      Towns: [1, 40, 75, 103, 109],
      None: 0,

      // Act 1
      RogueEncampment: 1,
      BloodMoor: 2,
      ColdPlains: 3,
      StonyField: 4,
      DarkWood: 5,
      BlackMarsh: 6,
      TamoeHighland: 7,
      DenofEvil: 8,
      CaveLvl1: 9,
      UndergroundPassageLvl1: 10,
      HoleLvl1: 11,
      PitLvl1: 12,
      CaveLvl2: 13,
      UndergroundPassageLvl2: 14,
      HoleLvl2: 15,
      PitLvl2: 16,
      BurialGrounds: 17,
      Crypt: 18,
      Mausoleum: 19,
      ForgottenTower: 20,
      TowerCellarLvl1: 21,
      TowerCellarLvl2: 22,
      TowerCellarLvl3: 23,
      TowerCellarLvl4: 24,
      TowerCellarLvl5: 25,
      MonasteryGate: 26,
      OuterCloister: 27,
      Barracks: 28,
      JailLvl1: 29,
      JailLvl2: 30,
      JailLvl3: 31,
      InnerCloister: 32,
      Cathedral: 33,
      CatacombsLvl1: 34,
      CatacombsLvl2: 35,
      CatacombsLvl3: 36,
      CatacombsLvl4: 37,
      Tristram: 38,
      MooMooFarm: 39,

      // Act 2
      LutGholein: 40,
      RockyWaste: 41,
      DryHills: 42,
      FarOasis: 43,
      LostCity: 44,
      ValleyofSnakes: 45,
      CanyonofMagic: 46,
      A2SewersLvl1: 47,
      A2SewersLvl2: 48,
      A2SewersLvl3: 49,
      HaremLvl1: 50,
      HaremLvl2: 51,
      PalaceCellarLvl1: 52,
      PalaceCellarLvl2: 53,
      PalaceCellarLvl3: 54,
      StonyTombLvl1: 55,
      HallsoftheDeadLvl1: 56,
      HallsoftheDeadLvl2: 57,
      ClawViperTempleLvl1: 58,
      StonyTombLvl2: 59,
      HallsoftheDeadLvl3: 60,
      ClawViperTempleLvl2: 61,
      MaggotLairLvl1: 62,
      MaggotLairLvl2: 63,
      MaggotLairLvl3: 64,
      AncientTunnels: 65,
      TalRashasTomb1: 66,
      TalRashasTomb2: 67,
      TalRashasTomb3: 68,
      TalRashasTomb4: 69,
      TalRashasTomb5: 70,
      TalRashasTomb6: 71,
      TalRashasTomb7: 72,
      DurielsLair: 73,
      ArcaneSanctuary: 74,

      // Act 3
      KurastDocktown: 75,
      SpiderForest: 76,
      GreatMarsh: 77,
      FlayerJungle: 78,
      LowerKurast: 79,
      KurastBazaar: 80,
      UpperKurast: 81,
      KurastCauseway: 82,
      Travincal: 83,
      SpiderCave: 84,
      SpiderCavern: 85,
      SwampyPitLvl1: 86,
      SwampyPitLvl2: 87,
      FlayerDungeonLvl1: 88,
      FlayerDungeonLvl2: 89,
      SwampyPitLvl3: 90,
      FlayerDungeonLvl3: 91,
      A3SewersLvl1: 92,
      A3SewersLvl2: 93,
      RuinedTemple: 94,
      DisusedFane: 95,
      ForgottenReliquary: 96,
      ForgottenTemple: 97,
      RuinedFane: 98,
      DisusedReliquary: 99,
      DuranceofHateLvl1: 100,
      DuranceofHateLvl2: 101,
      DuranceofHateLvl3: 102,

      // Act 4
      PandemoniumFortress: 103,
      OuterSteppes: 104,
      PlainsofDespair: 105,
      CityoftheDamned: 106,
      RiverofFlame: 107,
      ChaosSanctuary: 108,

      // Act 5
      Harrogath: 109,
      BloodyFoothills: 110,
      FrigidHighlands: 111,
      ArreatPlateau: 112,
      CrystalizedPassage: 113,
      FrozenRiver: 114,
      GlacialTrail: 115,
      DrifterCavern: 116,
      FrozenTundra: 117,
      AncientsWay: 118,
      IcyCellar: 119,
      ArreatSummit: 120,
      NihlathaksTemple: 121,
      HallsofAnguish: 122,
      HallsofPain: 123,
      HallsofVaught: 124,
      Abaddon: 125,
      PitofAcheron: 126,
      InfernalPit: 127,
      WorldstoneLvl1: 128,
      WorldstoneLvl2: 129,
      WorldstoneLvl3: 130,
      ThroneofDestruction: 131,
      WorldstoneChamber: 132,

      // Ubers
      MatronsDen: 133,
      ForgottenSands: 134,
      FurnaceofPain: 135,
      UberTristram: 136,

      actOf: function (area) {
        switch (true) {
        case area < sdk.areas.LutGholein:
          return 1;
        case area < sdk.areas.KurastDocktown:
          return 2;
        case area < sdk.areas.PandemoniumFortress:
          return 3;
        case area < sdk.areas.Harrogath:
          return 4;
        default:
          return 5;
        }
      },

      townOf: function (area) {
        switch (true) {
        case area < sdk.areas.LutGholein:
          return sdk.areas.RogueEncampment;
        case area < sdk.areas.KurastDocktown:
          return sdk.areas.LutGholein;
        case area < sdk.areas.PandemoniumFortress:
          return sdk.areas.KurastDocktown;
        case area < sdk.areas.Harrogath:
          return sdk.areas.PandemoniumFortress;
        default:
          return sdk.areas.Harrogath;
        }
      },

      townOfAct: function (act) {
        switch (act) {
        case 1:
          return sdk.areas.RogueEncampment;
        case 2:
          return sdk.areas.LutGholein;
        case 3:
          return sdk.areas.KurastDocktown;
        case 4:
          return sdk.areas.PandemoniumFortress;
        default:
          return sdk.areas.Harrogath;
        }
      }
    },

    skills: {
      get: {
        RightName: 0,
        LeftName: 1,
        RightId: 2,
        LeftId: 3,
        AllSkills: 4
      },
      hand: {
        Right: 0,
        Left: 1,
        LeftNoShift: 2,
        RightShift: 3,
      },
      subindex: {
        HardPoints: 0,
        SoftPoints: 1
      },
      // General
      Attack: 0,
      Kick: 1,
      Throw: 2,
      Unsummon: 3,
      LeftHandThrow: 4,
      LeftHandSwing: 5,

      // Amazon
      MagicArrow: 6,
      FireArrow: 7,
      InnerSight: 8,
      CriticalStrike: 9,
      Jab: 10,
      ColdArrow: 11,
      MultipleShot: 12,
      Dodge: 13,
      PowerStrike: 14,
      PoisonJavelin: 15,
      ExplodingArrow: 16,
      SlowMissiles: 17,
      Avoid: 18,
      Impale: 19,
      LightningBolt: 20,
      IceArrow: 21,
      GuidedArrow: 22,
      Penetrate: 23,
      ChargedStrike: 24,
      PlagueJavelin: 25,
      Strafe: 26,
      ImmolationArrow: 27,
      Dopplezon: 28,
      Decoy: 28,
      Evade: 29,
      Fend: 30,
      FreezingArrow: 31,
      Valkyrie: 32,
      Pierce: 33,
      LightningStrike: 34,
      LightningFury: 35,

      // Sorc
      FireBolt: 36,
      Warmth: 37,
      ChargedBolt: 38,
      IceBolt: 39,
      FrozenArmor: 40,
      Inferno: 41,
      StaticField: 42,
      Telekinesis: 43,
      FrostNova: 44,
      IceBlast: 45,
      Blaze: 46,
      FireBall: 47,
      Nova: 48,
      Lightning: 49,
      ShiverArmor: 50,
      FireWall: 51,
      Enchant: 52,
      ChainLightning: 53,
      Teleport: 54,
      GlacialSpike: 55,
      Meteor: 56,
      ThunderStorm: 57,
      EnergyShield: 58,
      Blizzard: 59,
      ChillingArmor: 60,
      FireMastery: 61,
      Hydra: 62,
      LightningMastery: 63,
      FrozenOrb: 64,
      ColdMastery: 65,

      // Necro
      AmplifyDamage: 66,
      Teeth: 67,
      BoneArmor: 68,
      SkeletonMastery: 69,
      RaiseSkeleton: 70,
      DimVision: 71,
      Weaken: 72,
      PoisonDagger: 73,
      CorpseExplosion: 74,
      ClayGolem: 75,
      IronMaiden: 76,
      Terror: 77,
      BoneWall: 78,
      GolemMastery: 79,
      RaiseSkeletalMage: 80,
      Confuse: 81,
      LifeTap: 82,
      PoisonExplosion: 83,
      BoneSpear: 84,
      BloodGolem: 85,
      Attract: 86,
      Decrepify: 87,
      BonePrison: 88,
      SummonResist: 89,
      IronGolem: 90,
      LowerResist: 91,
      PoisonNova: 92,
      BoneSpirit: 93,
      FireGolem: 94,
      Revive: 95,

      // Paladin
      Sacrifice: 96,
      Smite: 97,
      Might: 98,
      Prayer: 99,
      ResistFire: 100,
      HolyBolt: 101,
      HolyFire: 102,
      Thorns: 103,
      Defiance: 104,
      ResistCold: 105,
      Zeal: 106,
      Charge: 107,
      BlessedAim: 108,
      Cleansing: 109,
      ResistLightning: 110,
      Vengeance: 111,
      BlessedHammer: 112,
      Concentration: 113,
      HolyFreeze: 114,
      Vigor: 115,
      Conversion: 116,
      HolyShield: 117,
      HolyShock: 118,
      Sanctuary: 119,
      Meditation: 120,
      FistoftheHeavens: 121,
      Fanaticism: 122,
      Conviction: 123,
      Redemption: 124,
      Salvation: 125,

      // Barb
      Bash: 126,
      SwordMastery: 127,
      AxeMastery: 128,
      MaceMastery: 129,
      Howl: 130,
      FindPotion: 131,
      Leap: 132,
      DoubleSwing: 133,
      PoleArmMastery: 134,
      ThrowingMastery: 135,
      SpearMastery: 136,
      Taunt: 137,
      Shout: 138,
      Stun: 139,
      DoubleThrow: 140,
      IncreasedStamina: 141,
      FindItem: 142,
      LeapAttack: 143,
      Concentrate: 144,
      IronSkin: 145,
      BattleCry: 146,
      Frenzy: 147,
      IncreasedSpeed: 148,
      BattleOrders: 149,
      GrimWard: 150,
      Whirlwind: 151,
      Berserk: 152,
      NaturalResistance: 153,
      WarCry: 154,
      BattleCommand: 155,

      // General stuff
      IdentifyScroll: 217,
      BookofIdentify: 218,
      TownPortalScroll: 219,
      BookofTownPortal: 220,

      // Druid
      Raven: 221,
      PoisonCreeper: 222, // External
      PlaguePoppy: 222, // Internal
      Werewolf: 223, // External
      Wearwolf: 223, // Internal
      Lycanthropy: 224, // External
      ShapeShifting: 224, // Internal
      Firestorm: 225,
      OakSage: 226,
      SpiritWolf: 227, // External
      SummonSpiritWolf: 227, // Internal
      Werebear: 228, // External
      Wearbear: 228, // Internal
      MoltenBoulder: 229,
      ArcticBlast: 230,
      CarrionVine: 231, // External
      CycleofLife: 231, // Internal
      FeralRage: 232,
      Maul: 233,
      Fissure: 234, // Internal
      Eruption: 234, // Internal
      CycloneArmor: 235,
      HeartofWolverine: 236,
      SummonDireWolf: 237, // External
      SummonFenris: 237, // Internal
      Rabies: 238,
      FireClaws: 239,
      Twister: 240,
      SolarCreeper: 241, // External
      Vines: 241, // Internal
      Hunger: 242,
      ShockWave: 243,
      Volcano: 244,
      Tornado: 245,
      SpiritofBarbs: 246,
      Grizzly: 247, // External
      SummonGrizzly: 247, // Internal
      Fury: 248,
      Armageddon: 249,
      Hurricane: 250,

      // Assa
      FireBlast: 251, // External
      FireTrauma: 251, // Internal
      ClawMastery: 252,
      PsychicHammer: 253,
      TigerStrike: 254,
      DragonTalon: 255,
      ShockWeb: 256, // External
      ShockField: 256, // Internal
      BladeSentinel: 257,
      Quickness: 258, // Internal name
      BurstofSpeed: 258, // Shown name
      FistsofFire: 259,
      DragonClaw: 260,
      ChargedBoltSentry: 261,
      WakeofFire: 262, // External
      WakeofFireSentry: 262, // Internal
      WeaponBlock: 263,
      CloakofShadows: 264,
      CobraStrike: 265,
      BladeFury: 266,
      Fade: 267,
      ShadowWarrior: 268,
      ClawsofThunder: 269,
      DragonTail: 270,
      LightningSentry: 271,
      WakeofInferno: 272, // External
      InfernoSentry: 272, // Internal
      MindBlast: 273,
      BladesofIce: 274,
      DragonFlight: 275,
      DeathSentry: 276,
      BladeShield: 277,
      Venom: 278,
      ShadowMaster: 279,
      PhoenixStrike: 280, // External
      RoyalStrike: 280, // Internal
      WakeofDestructionSentry: 281, // Not used?
      Summoner: 500, // special
      tabs: {
        // Ama
        BowandCrossbow: 0,
        PassiveandMagic: 1,
        JavelinandSpear: 2,

        // Sorc
        Fire: 8,
        Lightning: 9,
        Cold: 10,

        // Necro
        Curses: 16,
        PoisonandBone: 17,
        NecroSummoning: 18,

        // Pala
        PalaCombat: 24,
        Offensive: 25,
        Defensive: 26,

        // Barb
        BarbCombat: 32,
        Masteries: 33,
        Warcries: 34,

        // Druid
        DruidSummon: 40,
        ShapeShifting: 41,
        Elemental: 42,

        // Assa
        Traps: 48,
        ShadowDisciplines: 49,
        MartialArts: 50,
      }
    },
    skillTabs: undefined,

    quest: {
      item: {
        // Act 1
        WirtsLeg: 88,
        HoradricMalus: 89,
        ScrollofInifuss: 524,
        KeytotheCairnStones: 525,
        // Act 2
        FinishedStaff: 91,
        "HoradricStaff": 91,
        IncompleteStaff: 92,
        "ShaftoftheHoradricStaff": 92,
        ViperAmulet: 521,
        "TopoftheHoradricStaff": 521,
        Cube: 549,
        BookofSkill: 552,
        // Act 3
        "DecoyGidbinn": 86,
        "TheGidbinn": 87,
        KhalimsFlail: 173,
        KhalimsWill: 174,
        PotofLife: 545,
        "AJadeFigurine": 546,
        JadeFigurine: 546,
        TheGoldenBird: 547,
        LamEsensTome: 548,
        KhalimsEye: 553,
        KhalimsHeart: 554,
        KhalimsBrain: 555,
        // Act 4
        HellForgeHammer: 90,
        Soulstone: 551,
        "MephistosSoulstone": 551,
        // Act 5
        MalahsPotion: 644,
        "ScrollofKnowledge": 645,
        ScrollofResistance: 646,
        // Pandemonium Event
        KeyofTerror: 647,
        KeyofHate: 648,
        KeyofDestruction: 649,
        DiablosHorn: 650,
        BaalsEye: 651,
        MephistosBrain: 652,
        StandardofHeroes: 658,
        // Essences/Token
        TokenofAbsolution: 653,
        TwistedEssenceofSuffering: 654,
        ChargedEssenceofHatred: 655,
        BurningEssenceofTerror: 656,
        FesteringEssenceofDestruction: 657,
        // Misc
        "TheBlackTowerKey": 544,
      },
      items: [
        // act 1
        88, 89, 524, 525,
        // act 2
        91, 92, 521, 549, 552,
        // act 3
        86, 87, 173, 174, 545, 546, 547, 548, 553, 554, 555,
        // act 4
        90, 551,
        // act 5
        644, 645, 646,
      ],
      chest: {
        // act1
        StoneAlpha: 17,
        StoneBeta: 18,
        StoneGamma: 19,
        StoneDelta: 20,
        StoneLambda: 21,
        StoneTheta: 22, // ?
        CainsJail: 26,
        InifussTree: 30,
        MalusHolder: 108,
        Wirt: 268,

        // act 2
        ViperAmuletChest: 149,
        HoradricStaffHolder: 152,
        HoradricCubeChest: 354,
        HoradricScrollChest: 355,
        ShaftoftheHoradricStaffChest: 356,
        Journal: 357,

        // act 3
        ForestAltar: 81,
        LamEsensTomeHolder: 193,
        GidbinnAltar: 252,
        KhalimsHeartChest: 405,
        KhalimsBrainChest: 406,
        KhalimsEyeChest: 407,

        // act 4
        HellForge: 376,

        // act 5
        BarbCage: 473,
        FrozenAnya: 558,
        AncientsAltar: 546,
      },
      chests: [
        // act 1
        17, 18, 19, 20, 21, 22, 26, 30, 108,
        // act 2
        149, 152, 354, 355, 356, 357,
        // act 3
        81, 193, 405, 406, 407,
        // act 4
        376,
        // act 5
        434, 558, 546
      ],
      id: {
        SpokeToWarriv: 0,
        DenofEvil: 1,
        SistersBurialGrounds: 2,
        TheSearchForCain: 4,
        ForgottenTower: 5,
        ToolsoftheTrade: 3,
        SistersToTheSlaughter: 6,
        AbleToGotoActII: 7,
        SpokeToJerhyn: 8,
        RadamentsLair: 9,
        TheHoradricStaff: 10,
        TheTaintedSun: 11,
        TheArcaneSanctuary: 12,
        TheSummoner: 13,
        TheSevenTombs: 14,
        AbleToGotoActIII: 15,
        SpokeToHratli: 16,
        TheGoldenBird: 20,
        BladeoftheOldReligion: 19,
        KhalimsWill: 18,
        LamEsensTome: 17,
        TheBlackenedTemple: 21,
        TheGuardian: 22,
        AbleToGotoActIV: 23,
        SpokeToTyrael: 24,
        TheFallenAngel: 25,
        HellsForge: 27,
        TerrorsEnd: 26,
        AbleToGotoActV: 28,
        SiegeOnHarrogath: 35,
        RescueonMountArreat: 36,
        PrisonofIce: 37,
        BetrayalofHarrogath: 38,
        RiteofPassage: 39,
        EyeofDestruction: 40,
        Respec: 41,
      },
      // just common states for now
      states: {
        Completed: 0,
        ReqComplete: 1,
        GreyedOut: 12,
        PartyMemberComplete: 13,
        CannotComplete: 14,
      }
    },

    // in game data
    uiflags: {
      Inventory: 0x01,
      StatsWindow: 0x02,
      QuickSkill: 0x03,
      SkillWindow: 0x04,
      ChatBox: 0x05,
      NPCMenu: 0x08,
      EscMenu: 0x09,
      KeytotheCairnStonesScreen: 0x10,
      AutoMap: 0x0A,
      ConfigControls: 0x0B,
      Shop: 0x0C,
      ShowItem: 0x0D,
      SubmitItem: 0x0E,
      Quest: 0x0F,
      QuestLog: 0x11,
      StatusArea: 0x12,
      Waypoint: 0x14,
      MiniPanel: 0x15,
      Party: 0x16,
      TradePrompt: 0x17,
      Msgs: 0x18,
      Stash: 0x19,
      Cube: 0x1A,
      ShowBelt: 0x1F,
      Help: 0x21,
      MercScreen: 0x24,
      ScrollWindow: 0x25
    },

    menu: {
      Respec: 0x2BA0,
      Ok: 0x0D49,
      Talk: 0x0D35,
      Trade: 0x0D44,
      TradeRepair: 0x0D06,
      Imbue: 0x0FB1,
      Gamble: 0x0D46,
      Hire: 0x0D45,
      GoEast: 0x0D36,
      GoWest: 0x0D37,
      IdentifyItems: 0x0FB4,
      SailEast: 0x0D38,
      SailWest: 0x0D39,
      RessurectMerc: 0x1507,
      AddSockets: 0x58DC,
      Personalize: 0x58DD,
      TravelToHarrogath: 0x58D2,
    },

    // shrine types
    shrines: {
      Presets: [2, 81, 83, 170, 344, 197, 202],
      Ids: [
        2, 77, 81, 83, 84, 85, 93, 96, 97, 109, 116, 123, 124,
        133, 134, 135, 136, 150, 151, 164, 165, 166, 167, 168,
        170, 172, 173, 184, 190, 191, 197, 199, 200, 201, 202,
        206, 226, 231, 232, 236, 249, 260, 262, 263, 264, 265,
        275, 276, 277, 278, 279, 280, 281, 282, 299, 300, 302,
        303, 320, 325, 343, 344, 361, 414, 415, 421, 422, 423,
        427, 428, 464, 465, 472, 479, 483, 484, 488, 491, 492,
        495, 497, 499, 503, 509, 512, 520, 521, 522
      ],
      None: 0,
      Refilling: 1,
      Health: 2,
      Mana: 3,
      HealthExchange: 4,
      ManaExchange: 5,
      Armor: 6,
      Combat: 7,
      ResistFire: 8,
      ResistCold: 9,
      ResistLightning: 10,
      ResistPoison: 11,
      Skill: 12,
      ManaRecharge: 13,
      Stamina: 14,
      Experience: 15,
      Enirhs: 16,
      Portal: 17,
      Gem: 18,
      Fire: 19,
      Monster: 20,
      Exploding: 21,
      Poison: 22
    },

    // unit states
    states: {
      None: 0,
      FrozenSolid: 1,
      Poison: 2,
      ResistFire: 3,
      ResistCold: 4,
      ResistLightning: 5,
      ResistMagic: 6,
      PlayerBody: 7,
      ResistAll: 8,
      AmplifyDamage: 9,
      FrozenArmor: 10,
      Frozen: 11,
      Inferno: 12,
      Blaze: 13,
      BoneArmor: 14,
      Concentrate: 15,
      Enchant: 16,
      InnerSight: 17,
      SkillMove: 18,
      Weaken: 19,
      ChillingArmor: 20,
      Stunned: 21,
      SpiderLay: 22,
      DimVision: 23,
      Slowed: 24,
      FetishAura: 25,
      Shout: 26,
      Taunt: 27,
      Conviction: 28,
      Convicted: 29,
      EnergyShield: 30,
      Venom: 31,
      BattleOrders: 32,
      Might: 33,
      Prayer: 34,
      HolyFire: 35,
      Thorns: 36,
      Defiance: 37,
      ThunderStorm: 38,
      LightningBolt: 39,
      BlessedAim: 40,
      Stamina: 41,
      Concentration: 42,
      Holywind: 43,
      HolyFreeze: 43,
      HolywindCold: 44,
      HolyFreezeCold: 44,
      Cleansing: 45,
      HolyShock: 46,
      Sanctuary: 47,
      Meditation: 48,
      Fanaticism: 49,
      Redemption: 50,
      BattleCommand: 51,
      PreventHeal: 52,
      Conversion: 53,
      Uninterruptable: 54,
      IronMaiden: 55,
      Terror: 56,
      Attract: 57,
      LifeTap: 58,
      Confuse: 59,
      Decrepify: 60,
      LowerResist: 61,
      OpenWounds: 62,
      Dopplezon: 63,
      Decoy: 63,
      CriticalStrike: 64,
      Dodge: 65,
      Avoid: 66,
      Penetrate: 67,
      Evade: 68,
      Pierce: 69,
      Warmth: 70,
      FireMastery: 71,
      LightningMastery: 72,
      ColdMastery: 73,
      SwordMastery: 74,
      AxeMastery: 75,
      MaceMastery: 76,
      PoleArmMastery: 77,
      ThrowingMastery: 78,
      SpearMastery: 79,
      IncreasedStamina: 80,
      IronSkin: 81,
      IncreasedSpeed: 82,
      NaturalResistance: 83,
      FingerMageCurse: 84,
      NoManaReg: 85,
      JustHit: 86,
      SlowMissiles: 87,
      ShiverArmor: 88,
      BattleCry: 89,
      Blue: 90,
      Red: 91,
      DeathDelay: 92,
      Valkyrie: 93,
      Frenzy: 94,
      Berserk: 95,
      Revive: 96,
      ItemFullSet: 97,
      SourceUnit: 98,
      Redeemed: 99,
      HealthPot: 100,
      HolyShield: 101,
      JustPortaled: 102,
      MonFrenzy: 103,
      CorpseNoDraw: 104,
      Alignment: 105,
      ManaPot: 106,
      Shatter: 107,
      SyncWarped: 108,
      ConversionSave: 109,
      Pregnat: 110,
      Rabies: 112,
      DefenceCurse: 113,
      BloodMana: 114,
      Burning: 115,
      DragonFlight: 116,
      Maul: 117,
      CorpseNoSelect: 118,
      ShadowWarrior: 119,
      FeralRage: 120,
      SkillDelay: 121,
      ProgressiveDamage: 122,
      ProgressiveSteal: 123,
      ProgressiveOther: 124,
      ProgressiveFire: 125,
      ProgressiveCold: 126,
      ProgressiveLighting: 127,
      ShrineArmor: 128,
      ShrineCombat: 129,
      ShrineResLighting: 130,
      ShrineResFire: 131,
      ShrineResCold: 132,
      ShrineResPoison: 133,
      ShrineSkill: 134,
      ShrineManaRegen: 135,
      ShrineStamina: 136,
      ShrineExperience: 137,
      FenrisRage: 138,
      Wolf: 139,
      Wearwolf: 139,
      Bear: 140,
      Wearbear: 140,
      Bloodlust: 141,
      ChangeClass: 142,
      Attached: 143,
      Hurricane: 144,
      Armageddon: 145,
      Invis: 146,
      Barbs: 147,
      HeartofWolverine: 148,
      OakSage: 149,
      VineBeast: 150,
      CycloneArmor: 151,
      ClawMastery: 152,
      CloakofShadows: 153,
      Recyled: 154,
      WeaponBlock: 155,
      Cloaked: 156,
      Quickness: 157, // Internal name
      BurstofSpeed: 157, // External name
      BladeShield: 158,
      Fade: 159,
      RestInPeace: 172,
      Glowing: 175,
      Delerium: 177,
      Antidote: 178,
      Thawing: 179,
      StaminaPot: 180,
    },

    enchant: {
      RandName: 1,
      HpMultiply: 2,
      AddLightRadius: 3,
      AddMLvl: 4,
      ExtraStrong: 5,
      ExtraFast: 6,
      Cursed: 7,
      MagicResistant: 8,
      FireEnchanted: 9,
      PoisonDeath: 10,
      InsectDeath: 11,
      ChainLightingDeath: 12,
      IgnoreTargetDefense: 13,
      UnknownMod: 14,
      KillMinionsDeath: 15,
      ChampMods: 16,
      LightningEnchanted: 17,
      ColdEnchanted: 18,
      UnusedMercMod: 19,
      ChargedBoltWhenStruck: 20,
      TempSummoned: 21,
      QuestMod: 22,
      PoisonField: 23,
      Thief: 24,
      ManaBurn: 25,
      Teleportation: 26,
      SpectralHit: 27,
      StoneSkin: 28,
      MultipleShots: 29,
      Aura: 30,
      CorpseExplosion: 31,
      FireExplosionOnDeath: 32, // not sure what the difference is between this and 9
      FreezeOnDeath: 33,
      SelfResurrect: 34,
      IceShatter: 35,
      ChampStoned: 36,
      ChampStats: 37,
      ChampCurseImmune: 38,
    },

    // unit stats
    stats: {
      StunLength: 66,
      VelocityPercent: 67,
      OtherAnimrate: 69,
      HpRegen: 74,

      LastBlockFrame: 95,
      State: 98,
      MonsterPlayerCount: 100,

      CurseResistance: 109,
      IronMaidenLevel: 129,
      LifeTapLevel: 130,

      Alignment: 172,
      Target0: 173,
      Target1: 174,
      GoldLost: 175,
      MinimumRequiredLevel: 176,
      ConversionLevel: 176,
      ConversionMaxHp: 177,
      UnitDooverlay: 178,
      AttackVsMontype: 179,
      DamageVsMontype: 180,

      ArmorOverridePercent: 182,
      FireLength: 315,
      BurningMin: 316,
      BurningMax: 317,
      ProgressiveDamage: 318,
      ProgressiveSteal: 319,
      ProgressiveOther: 320,
      ProgressiveFire: 321,
      ProgressiveCold: 322,
      ProgressiveLightning: 323,
      ProgressiveTohit: 325,
      PoisonCount: 326,
      DamageFramerate: 327,
      PierceIdx: 328,

      ModifierListSkill: 350,
      ModifierListLevel: 351,

      LastSentHpPct: 352,
      SourceUnitType: 353,
      SourceUnitId: 354,

      SkillThornsPercent: 131,
      SkillBoneArmor: 132,
      SkillCycloneArmor: 132,
      SkillBoneArmorMax: 133,
      SkillCycloneArmorMax: 133,
      SkillFade: 181,
      SkillPoisonOverrideLength: 101,
      SkillBypassUndead: 103,
      SkillBypassDemons: 104,
      SkillBypassBeasts: 106,
      SkillHandofAthena: 161,
      SkillStaminaPercent: 162,
      SkillPassiveStaminaPercent: 163,
      SkillConcentration: 164,
      SkillEnchant: 165,
      SkillPierce: 166,
      SkillConviction: 167,
      SkillChillingArmor: 168,
      SkillFrenzy: 169,
      SkillDecrepify: 170,
      SkillArmorPercent: 171,

      Strength: 0,
      Energy: 1,
      Dexterity: 2,
      Vitality: 3,
      StatPts: 4,
      NewSkills: 5,
      HitPoints: 6,
      MaxHp: 7,
      Mana: 8,
      MaxMana: 9,
      Stamina: 10,
      MaxStamina: 11,
      Level: 12,
      Experience: 13,
      Gold: 14,
      GoldBank: 15,
      ArmorPercent: 16,
      MaxDamagePercent: 17,
      MinDamagePercent: 18,
      EnhancedDamage: 18,
      ToHit: 19,
      ToBlock: 20,
      MinDamage: 21,
      MaxDamage: 22,
      SecondaryMinDamage: 23,
      SecondaryMaxDamage: 24,
      DamagePercent: 25,
      ManaRecovery: 26,
      ManaRecoveryBonus: 27,
      StaminaRecoveryBonus: 28,
      LastExp: 29,
      NextExp: 30,
      ArmorClass: 31,
      Defense: 31,
      ArmorClassVsMissile: 32,
      ArmorClassVsHth: 33,
      NormalDamageReduction: 34,
      MagicDamageReduction: 35,
      DamageResist: 36,
      MagicResist: 37,
      MaxMagicResist: 38,
      FireResist: 39,
      MaxFireResist: 40,
      LightResist: 41,
      LightningResist: 41,
      MaxLightResist: 42,
      ColdResist: 43,
      MaxColdResist: 44,
      PoisonResist: 45,
      MaxPoisonResist: 46,
      DamageAura: 47,
      FireMinDamage: 48,
      FireMaxDamage: 49,
      LightMinDamage: 50,
      LightMaxDamage: 51,
      MagicMinDamage: 52,
      MagicMaxDamage: 53,
      ColdMinDamage: 54,
      ColdMaxDamage: 55,
      ColdLength: 56,
      PoisonMinDamage: 57,
      PoisonMaxDamage: 58,
      PoisonLength: 59,
      LifeDrainMinDamage: 60,
      LifeLeech: 60,
      LifeDrainMaxDamage: 61,
      ManaDrainMinDamage: 62,
      ManaLeech: 62,
      ManaDrainMaxDamage: 63,
      StaminaDrainMinDamage: 64,
      StaminaDrainMaxDamage: 65,
      AttackRate: 68,
      PreviousSkillRight: 181,
      PreviousSkillMiddle: 182,
      PreviousSkillLeft: 183,
      PassiveFireMastery: 329,
      PassiveLightningMastery: 330,
      PassiveColdMastery: 331,
      PassivePoisonMastery: 332,
      PassiveFirePierce: 333,
      PassiveLightningPierce: 334,
      PassiveColdPierce: 335,
      PassivePoisonPierce: 336,
      PassiveCriticalStrike: 337,
      PassiveDodge: 338,
      PassiveAvoid: 339,
      PassiveEvade: 340,
      PassiveWarmth: 341,
      PassiveMasteryMeleeTh: 342,
      PassiveMasteryMeleeDmg: 343,
      PassiveMasteryMeleeCrit: 344,
      PassiveMasteryThrowTh: 345,
      PassiveMasteryThrowDmg: 346,
      PassiveMasteryThrowCrit: 347,
      PassiveWeaponBlock: 348,
      PassiveSummonResist: 349,
      PassiveMagMastery: 357,
      PassiveMagPierce: 358,
      Quantity: 70,
      Value: 71,
      Durability: 72,
      MaxDurability: 73,
      MaxDurabilityPercent: 75,
      MaxHpPercent: 76,
      MaxManaPercent: 77,
      AttackerTakesDamage: 78,
      GoldBonus: 79,
      MagicBonus: 80,
      Knockback: 81,
      TimeDuration: 82,
      AddClassSkills: 83,
      AddExperience: 85,
      HealAfterKill: 86,
      ReducedPrices: 87,
      DoubleHerbDuration: 88,
      LightRadius: 89,
      LightColor: 90,
      ReqPercent: 91,
      LevelReq: 92,
      FasterAttackRate: 93,
      IAS: 93,
      LevelReqPct: 94,
      FasterMoveVelocity: 96,
      FRW: 96,
      NonClassSkill: 97,
      OSkill: 97,
      FasterGetHitRate: 99,
      FHR: 99,
      FasterBlockRate: 102,
      FBR: 102,
      FasterCastRate: 105,
      FCR: 105,
      SingleSkill: 107,
      RestinPeace: 108,
      PoisonLengthResist: 110,
      NormalDamage: 111,
      Howl: 112,
      Stupidity: 113,
      DamagetoMana: 114,
      IgnoreTargetAc: 115,
      IgnoreTargetDefense: 115,
      FractionalTargetAc: 116,
      PreventHeal: 117,
      HalfFreezeDuration: 118,
      ToHitPercent: 119,
      DamageTargetAc: 120,
      DemonDamagePercent: 121,
      UndeadDamagePercent: 122,
      DemontoHit: 123,
      UndeadtoHit: 124,
      Throwable: 125,
      ElemSkill: 126,
      AllSkills: 127,
      AttackerTakesLightDamage: 128,
      Freeze: 134,
      OpenWounds: 135,
      CrushingBlow: 136,
      KickDamage: 137,
      ManaAfterKill: 138,
      HealAfterDemonKill: 139,
      ExtraBlood: 140,
      DeadlyStrike: 141,
      AbsorbFirePercent: 142,
      AbsorbFire: 143,
      AbsorbLightPercent: 144,
      AbsorbLight: 145,
      AbsorbMagicPercent: 146,
      AbsorbMagic: 147,
      AbsorbColdPercent: 148,
      AbsorbCold: 149,
      AbsorbSlash: 262,
      AbsorbCrush: 263,
      AbsorbThrust: 264,
      AbsorbSlashPercent: 265,
      AbsorbCrushPercent: 266,
      AbsorbThrustPercent: 267,
      Slow: 150,
      Indestructible: 152,
      CannotbeFrozen: 153,
      StaminaDrainPct: 154,
      Reanimate: 155,
      Pierce: 156,
      MagicArrow: 157,
      ExplosiveArrow: 158,
      ThrowMinDamage: 159,
      ThrowMaxDamage: 160,
      AddSkillTab: 188,
      NumSockets: 194,
      SkillOnAura: 151,
      SkillOnAttack: 195,
      SkillOnKill: 196,
      SkillOnDeath: 197,
      SkillOnHit: 198,
      SkillOnStrike: 198,
      SkillOnLevelUp: 199,
      SkillOnGetHit: 201,
      SkillWhenStruck: 201,
      ChargedSkill: 204,
      PerLevelArmor: 214,
      PerLevelArmorPercent: 215,
      PerLevelHp: 216,
      PerLevelMana: 217,
      PerLevelMaxDamage: 218,
      PerLevelMaxDamagePercent: 219,
      PerLevelStrength: 220,
      PerLevelDexterity: 221,
      PerLevelEnergy: 222,
      PerLevelVitality: 223,
      PerLevelTohit: 224,
      PerLevelTohitPercent: 225,
      PerLevelColdDamageMax: 226,
      PerLevelFireDamageMax: 227,
      PerLevelLtngDamageMax: 228,
      PerLevelPoisDamageMax: 229,
      PerLevelResistCold: 230,
      PerLevelResistFire: 231,
      PerLevelResistLtng: 232,
      PerLevelResistPois: 233,
      PerLevelAbsorbCold: 234,
      PerLevelAbsorbFire: 235,
      PerLevelAbsorbLtng: 236,
      PerLevelAbsorbPois: 237,
      PerLevelThorns: 238,
      PerLevelFindGold: 239,
      PerLevelFindMagic: 240,
      PerLevelRegenstamina: 241,
      PerLevelStamina: 242,
      PerLevelDamageDemon: 243,
      PerLevelDamageUndead: 244,
      PerLevelTohitDemon: 245,
      PerLevelTohitUndead: 246,
      PerLevelCrushingblow: 247,
      PerLevelOpenwounds: 248,
      PerLevelKickDamage: 249,
      PerLevelDeadlystrike: 250,
      PerLevelFindGems: 251,
      ReplenishDurability: 252,
      ReplenishQuantity: 253,
      ExtraStack: 254,
      Find: 255,
      SlashDamage: 256,
      SlashDamagePercent: 257,
      CrushDamage: 258,
      CrushDamagePercent: 259,
      ThrustDamage: 260,
      ThrustDamagePercent: 261,
      ArmorByTime: 268,
      ArmorPercentByTime: 269,
      HpByTime: 270,
      ManaByTime: 271,
      MaxDamageByTime: 272,
      MaxDamagePercentByTime: 273,
      StrengthByTime: 274,
      DexterityByTime: 275,
      EnergyByTime: 276,
      VitalityByTime: 277,
      TohitByTime: 278,
      TohitPercentByTime: 279,
      ColdDamageMaxByTime: 280,
      FireDamageMaxByTime: 281,
      LtngDamageMaxByTime: 282,
      PoisDamageMaxByTime: 283,
      ResistColdByTime: 284,
      ResistFireByTime: 285,
      ResistLtngByTime: 286,
      ResistPoisByTime: 287,
      AbsorbColdByTime: 288,
      AbsorbFireByTime: 289,
      AbsorbLtngByTime: 290,
      AbsorbPoisByTime: 291,
      FindGoldByTime: 292,
      FindMagicByTime: 293,
      RegenstaminaByTime: 294,
      StaminaByTime: 295,
      DamageDemonByTime: 296,
      DamageUndeadByTime: 297,
      TohitDemonByTime: 298,
      TohitUndeadByTime: 299,
      CrushingBlowByTime: 300,
      OpenWoundsByTime: 301,
      KickDamageByTime: 302,
      DeadlyStrikeByTime: 303,
      FindGemsByTime: 304,
      PierceCold: 305,
      PierceFire: 306,
      PierceLtng: 307,
      PiercePois: 308,
      DamageVsMonster: 309,
      DamagePercentVsMonster: 310,
      TohitVsMonster: 311,
      TohitPercentVsMonster: 312,
      AcVsMonster: 313,
      AcPercentVsMonster: 314,
      ExtraCharges: 324,
      QuestDifficulty: 356,

      // doesn't exist but define for prototypes
      AllRes: 555,
    },

    // unit info
    unittype: {
      Player: 0,
      NPC: 1,
      Monster: 1,
      Object: 2,
      Missile: 3,
      Item: 4,
      Stairs: 5, // ToDo: might be more as stairs
    },

    player: {
      flag: {
        Ignore: 2,
        Hostile: 8,
      },
      slot: {
        Main: 0,
        Secondary: 1
      },
      move: {
        Walk: 0,
        Run: 1
      },
      mode: { // sdk.player.mode.
        Death: 0,
        StandingOutsideTown: 1,
        Walking: 2,
        Running: 3,
        GettingHit: 4,
        StandingInTown: 5,
        WalkingInTown: 6,
        Attacking1: 7,
        Attacking2: 8,
        Blocking: 9,
        CastingSkill: 10,
        ThrowingItem: 11,
        Kicking: 12,
        UsingSkill1: 13,
        UsingSkill2: 14,
        UsingSkill3: 15,
        UsingSkill4: 16,
        Dead: 17,
        SkillActionSequence: 18,
        KnockedBack: 19,
      },
      class: {
        Amazon: 0,
        Sorceress: 1,
        Necromancer: 2,
        Paladin: 3,
        Barbarian: 4,
        Druid: 5,
        Assassin: 6,

        nameOf: function (classid) {
          if (classid === undefined || typeof classid !== "number") return false;
          if (classid < 0 || classid > 6) return false;
          return ["Amazon", "Sorceress", "Necromancer", "Paladin", "Barbarian", "Druid", "Assassin"][classid];
        }
      }
    },

    npcs: {
      // same as monsters but more clear to use units.npcs.mode
      mode: {
        Death: 0,
        Standing: 1,
        Walking: 2,
        GettingHit: 3,
        Attacking1: 4,
        Attacking2: 5,
        Blocking: 6,
        CastingSkill: 7,
        UsingSkill1: 8,
        UsingSkill2: 9,
        UsingSkill3: 10,
        UsingSkill4: 11,
        Dead: 12,
        KnockedBack: 13,
        Spawning: 14,
        Running: 15
      },

      Akara: 148,
      Alkor: 254,
      Asheara: 252,
      WarrivAct1: 155,
      WarrivAct2: 175,
      Atma: 176,
      Tyrael: 367,
      Tyrael2: 251,
      Tyrael3: 521,
      Charsi: 154,
      DeckardCain1: 146,
      DeckardCain2: 244,
      DeckardCain3: 245,
      DeckardCain4: 246,
      DeckardCain5: 265,
      DeckardCain6: 520,
      Drognan: 177,
      Elzix: 199,
      Fara: 178,
      Gheed: 147,
      Greiz: 198,
      Halbu: 257,
      Hratli: 253,
      Jamella: 405,
      Jerhyn: 201,
      Kaelan: 331,
      Kashya: 150,
      Larzuk: 511,
      Lysander: 202,
      Malah: 513,
      Meshif: 210,
      Meshif2: 264,
      Natalya: 297,
      Ormus: 255,
      NihlathakNPC: 526,
      Qualkehk: 515,
      RogueScout: 270,
      TempleGuard1: 52,
      TempleGuard2: 665,
      TempleGuard3: 666,
      Townguard1: 535,
      Townguard2: 536,
    },

    objects: {
      mode: {
        Inactive: 0,
        Interacted: 1,
        Active: 2,
      },

      chestIds: [
        5, 6, 87, 104, 105, 106, 107, 143, 140, 141, 144, 146, 147, 148, 176, 177, 181, 183, 198, 240, 241,
        242, 243, 329, 330, 331, 332, 333, 334, 335, 336, 354, 355, 356, 371, 387, 389, 390, 391, 397, 405,
        406, 407, 413, 420, 424, 425, 430, 431, 432, 433, 454, 455, 501, 502, 504, 505, 580, 581
      ],

      // act1
      MoldyTome: 8,
      A1TownFire: 39,
      A1Waypoint: 119,
      StoneAlpha: 17,
      StoneBeta: 18,
      StoneGamma: 19,
      StoneDelta: 20,
      StoneLambda: 21,
      StoneTheta: 22,
      CainsJail: 26,
      InifussTree: 30,
      Malus: 108,

      // act 2
      A2Waypoint: 156,
      A2UndergroundUpStairs: 22,
      TrapDoorA2: 74, // ancienttunnel/sewers act 2
      DoorbyDockAct2: 75, // incorrect ? TODO: figure out what 75 really corresponds to since the door is obj type 5 with classid 20
      PortaltoDurielsLair: 100,
      HoradricStaffHolder: 152,
      ArcaneSanctuaryPortal: 298,
      HoradricCubeChest: 354,
      HoradricScrollChest: 355,
      Journal: 357,

      // act 3
      A3Waypoint: 237,
      ForestAltar: 81,
      LamEsensTome: 193,
      SewerStairsA3: 366,
      SewerLever: 367,
      DuranceEntryStairs: 386,
      RedPortalToAct4: 342,
      CompellingOrb: 404,

      // act 4
      A4Waypoint: 398,
      SealGlow: 131,
      DiabloStar: 255,
      DiabloSealInfector: 392,
      DiabloSealInfector2: 393,
      DiabloSealSeis: 394,
      DiabloSealVizier: 396,
      DiabloSealVizier2: 395,
      RedPortalToAct5: 566, // The one of tyreal

      // act 5
      A5Waypoint: 429,
      SideCavesA5: 75, // FrozenRiver, DrifterCavern, IcyCellar
      Act5Gate: 449,
      KorlictheProtectorStatue: 474,
      TalictheDefenderStatue: 475,
      MadawctheGuardianStatue: 476,
      AncientsAltar: 546,
      ArreatEnterAncientsWay: 564,
      ArreatEnterWorldstone: 547,
      //AncientsDoor: 547,
      AncientsDoor: 547, // Worldstone keep lvl 1
      FrozenAnya: 558,
      FrozenAnyasPlatform: 460,
      NihlathaksPlatform: 462,
      WorldstonePortal: 563,

      FrigidHighlandsChest: 455,
      IcyCellarChest: 397,

      SmallSparklyChest: 397,
      LargeSparklyChest: 455,
      SuperChest: 580,

      // misc
      BubblingPoolofBlood: 82,
      HornShrine: 83,
      Stash: 267,
      BluePortal: 59,
      RedPortal: 60,
      Smoke: 401,
    },

    exits: {
      type: {
        WalkThrough: 1,
        Stairs: 2,
        RedPortal: 60,
      },
      preset: {
        AreaEntrance: 0, // special
        // act 1
        CaveHoleUp: 4,
        CaveHoleLvl2: 5,
        Crypt: 6,
        Mausoleum: 7,
        CryptMausExit: 8,
        JailUpStairs: 13,
        JailDownStairs: 14,
        CathedralDownStairs: 15,
        CathedralUpStairs: 16,
        CatacombsUpStairs: 17,
        CatacombsDownStairs: 18,

        // act 2
        A2SewersTrapDoor: 19,
        A2EnterSewersDoor: 20,
        A2ExitSewersDoor: 21,
        A2UndergroundUpStairs: 22,
        A2DownStairs: 23,
        EnterHaremStairs: 24,
        ExitHaremStairs: 25,
        PreviousLevelHaremRight: 26,
        PreviousLevelHaremLeft: 27,
        NextLevelHaremRight: 28,
        NextLevelHaremLeft: 29,
        PreviousPalaceRight: 30,
        PreviousPalaceLeft: 31,
        NextLevelPalace: 32,
        EnterStonyTomb: 33,
        EnterHalls: 36,
        EnterTalTomb1: 38,
        EnterTalTomb2: 39,
        EnterTalTomb3: 40,
        EnterTalTomb4: 41,
        EnterTalTomb5: 42,
        EnterTalTomb6: 43,
        EnterTalTomb7: 44,
        PreviousAreaTomb: 45,
        NextLevelTomb: 46,
        EnterMaggotLair: 47,
        PreviousAreaMaggotLair: 48,
        NextLevelMaggotLair: 49,
        AncientTunnelsTrapDoor: 50,
        EntrancetoDurielsLair: 100,

        // act 3
        EnterSpiderHole: 51,
        ExitSpiderHole: 52,
        EnterPit: 53,
        EnterDungeon: 54,
        PreviousAreaDungeon: 55,
        NextLevelDungeon: 56,
        A3EnterSewers: 57,
        A3ExitSewersUpperK: 58,
        A3SewersPreviousArea: 58,
        A3ExitSewers: 59,
        A3NextLevelSewers: 60,
        EnterTemple: 61,
        ExitTemple: 63,
        EnterDurance: 64,
        PreviousLevelDurance: 65,
        NextLevelDurance: 68,
        SewerStairsA3: 366,
        DuranceEntryStairs: 386,

        // act 4
        EnterRiverStairs: 69,
        ExitRiverStairs: 70,
        // act 5
        EnterCrystal: 71,
        A5ExitCave: 73,
        A5NextLevelCave: 74,
        EnterSubLevelCave: 75,
        EnterNithsTemple: 76,
        PreviousAreaNithsTemple: 77,
        NextAreaNithsTemple: 78,
        ArreatEnterAncientsWay: 79,
        ArreatEnterWorldstone: 80,
        PreviousAreaWorldstone: 81,
        NextAreaWorldstone: 82,
      },
    },
        
    monsters: {
      preset: {
        // Confirmed
        Izual: 256,
        Bishibosh: 734,
        Bonebreak: 735,
        Coldcrow: 736,
        Rakanishu: 737,
        TreeheadWoodFist: 738,
        Griswold: 739,
        TheCountess: 740,
        PitspawnFouldog: 741,
        FlamespiketheCrawler: 742,
        BoneAsh: 743,
        Radament: 744,
        BloodwitchtheWild: 745,
        Fangskin: 746,
        Beetleburst: 747,
        CreepingFeature: 748,
        ColdwormtheBurrower: 749,
        FireEye: 750,
        DarkElder: 751,
        TheSummoner: 752,
        AncientKaatheSoulless: 753,
        TheSmith: 754,
        SszarktheBurning: 755,
        WitchDoctorEndugu: 756,
        Stormtree: 757,
        BattlemaidSarina: 758,
        IcehawkRiftwing: 759,
        IsmailVilehand: 760,
        GelebFlamefinger: 761,
        BremmSparkfist: 762,
        ToorcIcefist: 763,
        WyandVoidfinger: 764,
        MafferDragonhand: 765,
        WingedDeath: 766,
        Taintbreeder: 768,
        RiftwraiththeCannibal: 769,
        InfectorofSouls: 770,
        LordDeSeis: 771,
        GrandVizierofChaos: 772,
        TheCowKing: 773,
        Corpsefire: 774,
        Hephasto: 775,
        ShenktheOverseer: 776,
        TalictheDefender: 777,
        MadawctheGuardian: 778,
        KorlictheProtector: 779,
        AxeDweller: 780,
        BonesawBreaker: 781,
        DacFarren: 782,
        EldritchtheRectifier: 783,
        EyebacktheUnleashed: 784,
        ThreshSocket: 785,
        Pindleskin: 786,
        SnapchipShatter: 787,
        AnodizedElite: 788,
        VinvearMolech: 789,
        SharpToothSayer: 790,
        MagmaTorquer: 791,
        BlazeRipper: 792,
        Frozenstein: 793,
        Nihlathak: 794,
        ColenzotheAnnihilator: 795,
        AchmeltheCursed: 796,
        BartuctheBloody: 797,
        VentartheUnholy: 798,
        ListertheTormentor: 799,
        BloodRaven: 805,

        // Unconfirmed
        // Questionable
        GriefGrumble: 741, // JailLvl2
        UniqueJailLvl3: 273,
        UniqueArcaneSanctuary: 371,
      },
      mode: {
        Death: 0,
        Standing: 1,
        Walking: 2,
        GettingHit: 3,
        Attacking1: 4,
        Attacking2: 5,
        Blocking: 6,
        CastingSkill: 7,
        UsingSkill1: 8,
        UsingSkill2: 9,
        UsingSkill3: 10,
        UsingSkill4: 11,
        Dead: 12,
        KnockedBack: 13,
        Spawning: 14,
        Running: 15
      },
      spectype: {
        All: 0,
        Super: 1,
        Champion: 2,
        Unique: 4,
        SuperUnique: 5,
        Magic: 6,
        Minion: 8,
      },
      // todo - determine what all these correlate to
      type: {
        Undead: 1,
        Demon: 2,
        Insect: 3,
        Human: 4,
        Construct: 5,
        LowUndead: 6,
        HighUndead: 7,
        Skeleton: 8,
        Zombie: 9,
        BigHead: 10,
        FoulCrow: 11,
        Fallen: 12,
        Brute: 13,
        SandRaider: 14,
        Wraith: 15,
        CorruptRogue: 16,
        Baboon: 17,
        GoatMan: 18,
        QuillRat: 19,
        SandMaggot: 20,
        Viper: 21,
        SandLeaper: 22,
        PantherWoman: 23,
        Swarm: 24,
        Scarab: 25,
        Mummy: 26,
        Unraveler: 27,
        Vulture: 28,
        Mosquito: 29,
        WillowWisp: 30,
        Arach: 31,
        ThornHulk: 32,
        Vampire: 33,
        BatDemon: 34,
        Fetish: 35,
        Blunderbore: 36,
        UndeadFetish: 37,
        Zakarum: 38,
        FrogDemon: 39,
        Tentacle: 40,
        FingerMage: 41,
        Golem: 42,
        Vilekind: 43,
        Regurgitator: 44,
        DoomKnight: 45,
        CouncilMember: 46,
        MegaDemon: 47,
        Bovine: 48,
        SeigeBeast: 49,
        SnowYeti: 50,
        Minion: 51,
        Succubus: 52,
        Overseer: 53,
        Imp: 54,
        FrozenHorror: 55,
        BloodLord: 56,
        DeathMauler: 57,
        PutridDefiler: 58,
      },
      DiablosBoneCage: 340,
      DiablosBoneCage2: 342,
      Dummy1: 149,
      Dummy2: 268,
      AbyssKnight: 311,
      Afflicted: 10,
      Afflicted2: 580,
      AlbinoRoach: 95,
      Ancient1: 104,
      Ancient2: 669,
      Ancient3: 670,
      Apparition: 41,
      Arach1: 122,
      Arach2: 685,
      Assailant: 33,
      Assailant2: 603,
      BaalColdMage: 381,
      Balrog1: 360,
      Balrog2: 686,
      Banished: 135,
      Barbs: 422,
      Bear1: 428,
      Bear2: 431,
      Beast: 441,
      BerserkSlayer: 462,
      BlackArcher: 163,
      BlackLancer1: 168,
      BlackLancer2: 617,
      BlackLocusts: 88,
      BlackRaptor1: 17,
      BlackRaptor2: 592,
      BlackRogue: 46,
      BlackSoul1: 121,
      BlackSoul2: 640,
      BlackVultureNest: 208,
      BloodBoss: 482,
      BloodBringer: 443,
      BloodClan1: 55,
      BloodClan2: 588,
      BloodDiver: 139,
      BloodGolem: 290,
      BloodHawk1: 16,
      BloodHawk2: 591,
      BloodHawkNest: 207,
      BloodHook: 116,
      BloodHookNest: 336,
      BloodLord1: 134,
      BloodLord2: 695,
      BloodWing: 117,
      BloodWingNest: 337,
      Blunderbore1: 186,
      Blunderbore2: 618,
      BoneArcher1: 172,
      BoneArcher2: 576,
      BoneMage1: 275,
      BoneMage2: 380,
      BoneMage3: 384,
      BoneMage4: 388,
      BoneMage5: 624,
      BoneWarrior1: 2,
      BoneWarrior2: 648,
      HellBovine: 391,
      BrambleHulk: 128,
      Brute: 24,
      Bunny: 556,
      BurningDead: 3,
      BurningDeadArcher1: 173,
      BurningDeadArcher2: 575,
      BurningDeadArcher3: 577,
      BurningDeadMage1: 276,
      BurningDeadMage2: 385,
      BurningDeadMage3: 389,
      BurningDeadMage4: 621,
      BurningSoul1: 641,
      BurningSoul2: 120,
      Cadaver1: 100,
      Cadaver2: 703,
      Cantor: 239,
      CarrionBird1: 110,
      CarrionBird2: 608,
      Carver1: 642,
      Carver2: 20,
      CarverShaman: 645,
      CarverShaman2: 59,
      CaveLeaper1: 79,
      CaveLeaper2: 629,
      ClawViper1: 74,
      ClawViper2: 594,
      CloudStalker1: 18,
      CloudStalker2: 593,
      CloudStalkerNest: 209,
      Combatant1: 522,
      Combatant2: 523,
      ConsumedFireBoar: 464,
      ConsumedIceBoar: 463,
      CorpseSpitter: 308,
      Corpulent: 307,
      Creature1: 248,
      Creature2: 427,
      Creeper: 413,
      CrushBiest: 442,
      Crusher: 26,
      Damned1: 14,
      Damned2: 584,
      DarkArcher1: 162,
      DarkArcher2: 614,
      DarkFamiliar: 140,
      DarkHunter: 43,
      DarkLancer1: 167,
      DarkLancer2: 616,
      DarkLord1: 133,
      DarkLord2: 697,
      DarkOne1: 22,
      DarkOne2: 644,
      DarkRanger: 160,
      DarkShaman1: 61,
      DarkShaman2: 647,
      DarkShape: 42,
      DarkSpearwoman: 165,
      DarkStalker: 45,
      DeamonSteed: 445,
      DeathClan1: 57,
      DeathClan2: 589,
      Decayed: 97,
      DefiledWarrior: 440,
      Defiler1: 546,
      Defiler2: 547,
      Defiler3: 548,
      Defiler4: 549,
      Defiler5: 550,
      DesertWing: 136,
      Destruction: 410,
      Devilkin: 643,
      Devilkin2: 21,
      DevilkinShaman: 646,
      DevilkinShaman2: 60,
      Devourer: 70,
      DevourerEgg: 192,
      DevourerQueen: 286,
      DevourerYoung: 182,
      Disfigured: 13,
      Disfigured2: 583,
      Dominus1: 474,
      Dominus2: 636,
      DoomApe: 51,
      DoomKnight: 310,
      DoomKnight1: 699,
      DoomKnight2: 700,
      Drehya1: 512,
      Drehya2: 527,
      DriedCorpse: 96,
      DrownedCarcass: 8,
      DuneBeast: 48,
      DungSoldier: 91,
      Dweller: 247,
      Eagle: 429,
      Embalmed: 98,
      Faithful: 236,
      Fallen: 19,
      FallenShaman: 58,
      FanaticMinion: 461,
      Feeder: 115,
      FeederNest: 335,
      Fenris: 421,
      Fetish1: 142,
      BoneFetish2: 213,
      Fetish3: 397,
      FetishShaman: 279,
      Fiend1: 137,
      Fiend2: 651,
      FireBoar: 456,
      FireTower: 372,
      FlameSpider: 125,
      Flayer1: 143,
      BoneFetish3: 214,
      Flayer3: 398,
      Flayer4: 659,
      Flayer5: 656,
      FlayerShaman1: 280,
      FlayerShaman2: 662,
      FleshArcher: 164,
      FleshBeast1: 301,
      FleshBeast2: 678,
      FleshHunter: 47,
      FleshLancer: 169,
      FleshSpawner1: 298,
      FleshSpawner2: 676,
      FlyingScimitar: 234,
      FoulCrow: 15,
      FoulCrow2: 590,
      FoulCrowNest: 206,
      FrenziedHellSpawn: 465,
      FrenziedIceSpawn: 466,
      GargantuanBeast: 28,
      Geglash: 200,
      Ghost1: 38,
      Ghost2: 631,
      Ghoul: 7,
      GhoulLord1: 131,
      GhoulLord2: 696,
      GiantLamprey: 71,
      GiantLampreyEgg: 193,
      GiantLampreyQueen: 287,
      GiantLampreyYoung: 183,
      GiantUrchin: 317,
      Gloam1: 118,
      Gloam2: 639,
      Gloombat1: 138,
      Gloombat2: 650,
      Gorbelly: 187,
      GoreBearer: 444,
      GreaterHellSpawn1: 459,
      GreaterHellSpawn2: 684,
      GreaterIceSpawn: 460,
      Groper: 304,
      Grotesque1: 300,
      Grotesque2: 675,
      GrotesqueWyrm1: 303,
      GrotesqueWyrm2: 677,
      Guardian1: 102,
      Guardian2: 667,
      Hawk: 419,
      Heirophant1: 240,
      Heirophant2: 241,
      Heirophant3: 673,
      Heirophant4: 674,
      HellBuzzard: 112,
      HellCat: 86,
      HellClan1: 56,
      HellClan2: 587,
      HellSlinger: 376,
      HellSpawn1: 457,
      HellSpawn2: 683,
      HellSwarm: 90,
      HellWhip: 483,
      HollowOne: 101,
      Horror: 4,
      Horror1: 501,
      Horror2: 502,
      Horror3: 503,
      Horror4: 504,
      Horror5: 505,
      HorrorArcher1: 174,
      HorrorArcher2: 579,
      HorrorMage1: 277,
      HorrorMage2: 382,
      HorrorMage3: 386,
      HorrorMage4: 390,
      HorrorMage5: 623,
      HorrorMage6: 625,
      HorrorMage7: 626,
      Hs1: 560,
      HungryDead: 6,
      Huntress1: 83,
      Huntress2: 627,
      Hut: 528,
      Hydra1: 351,
      Hydra2: 352,
      Hydra3: 353,
      IceBoar: 455,
      IceSpawn: 458,
      Imp1: 492,
      Imp2: 493,
      Imp3: 494,
      Imp4: 495,
      Imp5: 496,
      Imp6: 688,
      Imp7: 689,
      Infidel1: 32,
      Infidel2: 600,
      InsaneHellSpawn: 467,
      InsaneIceSpawn: 468,
      Invader1: 31,
      Invader2: 602,
      Itchies: 87,
      JungleHunter: 50,
      JungleUrchin: 67,
      Larva: 283,
      Lasher: 480,
      LightningSpire: 371,
      Lord1: 506,
      Lord2: 507,
      Lord3: 508,
      Lord4: 509,
      Lord5: 510,
      Lord6: 652,
      Lord7: 653,
      Maggot: 227,
      Malachai: 408,
      Marauder: 30,
      Marauder2: 599,
      Master: 418,
      Mauler: 188,
      Mauler1: 529,
      Mauler12: 604,
      Mauler2: 530,
      Mauler3: 531,
      Mauler4: 532,
      Mauler5: 533,
      Mauler6: 619,
      MawFiend: 694,
      MawFiend2: 309,
      Council1: 345,
      Council2: 346,
      Council3: 347,
      Council4: 557,
      Minion1: 572,
      Minion2: 573,
      Enslaved: 453,
      MinionSlayerSpawner: 485,
      MinionSpawner: 484,
      Misshapen1: 12,
      Misshapen2: 582,
      MoonClan1: 53,
      MoonClan2: 585,
      BaalSubjectMummy: 105,
      Navi: 266,
      Flavie: 266,
      NightClan1: 54,
      NightClan2: 586,
      NightLord: 132,
      NightMarauder: 295,
      NightSlinger1: 375,
      NightSlinger2: 395,
      NightTiger: 85,
      OblivionKnight1: 312,
      OblivionKnight2: 701,
      OblivionKnight3: 702,
      OverLord: 481,
      OverSeer: 479,
      PitLord1: 361,
      PitLord2: 687,
      PitViper1: 76,
      PitViper2: 595,
      PlagueBearer: 9,
      PlagueBugs: 89,
      PoisonSpinner: 124,
      PreservedDead: 99,
      ProwlingDead: 438,
      QuillBear: 313,
      QuillRat1: 63,
      QuillRat2: 605,
      RatMan1: 141,
      RatMan2: 396,
      BoneFetish1: 212,
      RatMan4: 407,
      RatManShaman: 278,
      RazorBeast: 316,
      RazorPitDemon: 82,
      RazorSpine1: 66,
      RazorSpine2: 607,
      ReanimatedHorde: 437,
      Returned1: 1,
      Returned2: 649,
      ReturnedArcher1: 171,
      ReturnedArcher2: 578,
      ReturnedMage: 274,
      ReturnedMage1: 379,
      ReturnedMage2: 383,
      ReturnedMage3: 387,
      ReturnedMage4: 620,
      ReturnedMage5: 622,
      RiverStalkerHead: 262,
      RiverStalkerLimb: 259,
      RockDweller: 49,
      RockWorm: 69,
      RockWormEgg: 191,
      RockWormQueen: 285,
      RockWormYoung: 181,
      RotWalker: 436,
      SaberCat1: 84,
      SaberCat2: 628,
      Salamander1: 75,
      Salamander2: 596,
      SandFisher: 123,
      SandLeaper: 78,
      SandMaggot: 68,
      SandMaggotEgg: 190,
      SandMaggotYoung: 180,
      SandRaider1: 29,
      SandRaider2: 601,
      DeathBeetle: 92,
      Scarab1: 93,
      Scarab2: 654,
      Sentry1: 411,
      Sentry2: 412,
      Sentry3: 415,
      Sentry4: 416,
      SerpentMagus1: 77,
      SerpentMagus2: 598,
      Sexton: 238,
      Skeleton: 0,
      SkeletonArcher: 170,
      Slayerexp1: 454,
      Slayerexp2: 682,
      Slinger1: 373,
      Slinger2: 610,
      Slinger3: 611,
      Slinger4: 612,
      SnowYeti1: 446,
      SnowYeti2: 447,
      SnowYeti3: 448,
      SnowYeti4: 449,
      SoulKiller: 691,
      SoulKiller1: 399,
      SoulKiller2: 144,
      SoulKiller3: 215,
      SoulKiller4: 658,
      SoulKiller5: 661,
      SoulKillerShaman1: 664,
      SoulKillerShaman2: 281,
      SpearCat: 394,
      SpearCat1: 374,
      Specter1: 40,
      Specter2: 633,
      SpiderMagus: 126,
      SpikeFiend1: 64,
      SpikeFiend2: 606,
      Spikefist: 130,
      SpikeGiant: 314,
      SteelWeevil1: 94,
      SteelWeevil2: 655,
      StormCaster1: 306,
      StormCaster2: 693,
      Strangler1: 305,
      Strangler2: 692,
      StygianDog: 302,
      StygianDoll1: 145,
      StygianDoll2: 216,
      StygianDoll3: 400,
      StygianDoll4: 660,
      StygianDoll5: 657,
      StygianDoll6: 690,
      StygianDollShaman1: 663,
      StygianDollShaman2: 282,
      StygianFury: 476,
      StygianHag: 299,
      StygianHarlot: 471,
      StygianWatcherHead: 263,
      StygianWatcherLimb: 260,
      Succubusexp1: 469,
      Succubusexp2: 634,
      Sucker: 114,
      SuckerNest: 334,
      Summoner: 250,
      SwampGhost: 119,
      Tainted: 11,
      Tainted2: 581,
      Taunt: 545,
      Temptress1: 472,
      Temptress2: 473,
      Temptress3: 635,
      Tentacle1: 562,
      Tentacle2: 563,
      Tentacle3: 564,
      Tentacle4: 565,
      Tentacle5: 566,
      ThornBeast: 65,
      ThornBrute: 315,
      ThornedHulk1: 127,
      ThornedHulk2: 609,
      Thrasher: 129,
      TombCreeper1: 80,
      TombCreeper2: 630,
      TombViper1: 73,
      TombViper2: 597,
      TrappedSoul1: 403,
      TrappedSoul2: 404,
      TreeLurker: 81,
      UndeadScavenger: 111,
      UnholyCorpse1: 439,
      UnholyCorpse2: 698,
      Unraveler1: 103,
      Unraveler2: 668,
      Urdar: 189,
      VenomLord1: 362,
      VenomLord2: 558,
      VileArcher1: 161,
      VileArcher2: 613,
      VileHunter: 44,
      VileLancer1: 166,
      VileLancer2: 615,
      VileTemptress: 470,
      VileWitch1: 475,
      VileWitch2: 638,
      WailingBeast: 27,
      WarpedFallen: 23,
      WarpedShaman: 62,
      Warrior: 417,
      WaterWatcherHead: 261,
      WaterWatcherLimb: 258,
      WingedNightmare: 113,
      Witch1: 637,
      Witch2: 477,
      Witch3: 478,
      Wolf1: 359,
      Wolf2: 420,
      Wolf3: 430,
      WolfRider1: 450,
      WolfRider2: 451,
      WolfRider3: 452,
      WorldKiller1: 679,
      WorldKiller2: 72,
      WorldKillerEgg1: 681,
      WorldKillerEgg2: 194,
      WorldKillerQueen: 288,
      WorldKillerYoung1: 680,
      WorldKillerYoung2: 184,
      Worm1: 551,
      Worm2: 552,
      Worm3: 553,
      Worm4: 554,
      Worm5: 555,
      Wraith1: 39,
      Wraith2: 632,
      Yeti: 25,
      Zakarumite: 235,
      Zealot1: 237,
      Zealot2: 671,
      Zealot3: 672,
      Zombie: 5,

      // Bosses/Ubers
      Andariel: 156,
      Duriel: 211,
      Mephisto: 242,
      Diablo: 243,
      DiabloClone: 333,
      ThroneBaal: 543,
      Baal: 544,
      BaalClone: 570,
      UberMephisto: 704,
      UberBaal: 705,
      UberIzual: 706,
      Lilith: 707,
      UberDuriel: 708,
      UberDiablo: 709,

      // Mini-Bosses
      TheSmith: 402,
      BloodRaven: 267,
      Radament: 229,
      TheSummoner: 250,
      Griswold: 365,
      Izual: 256,
      Hephasto: 409,
      KorlictheProtector: 540,
      TalictheDefender: 541,
      MadawctheGuardian: 542,
      ListerTheTormenter: 571,
      TheCowKing: 743, // 773?
      ColdwormtheBurrower: 284,
      Nihlathak: 526,

      // Objects
      Turret1: 348,
      Turret2: 349,
      Turret3: 350,
      CatapultS: 497,
      CatapultE: 498,
      CatapultSiege: 499,
      CatapultW: 500,
      Compellingorb: 366,
      GargoyleTrap: 273,
      MummyGenerator: 228,
      Stairs: 559,
      BarricadeDoor1: 432,
      BarricadeDoor2: 433,
      PrisonDoor: 434,
      BarricadeTower: 435,
      BarricadeWall1: 524,
      BarricadeWall2: 525,

      // Misc?
      Youngdiablo: 368,
      Left: 525,
      Life: 426,
      Effect: 574,
      Pet: 414,
      Prince: 249,
      POW: 534,
      Right: 524,
      Sage: 424,
      Town: 514,
      Cow: 179,
    },

    summons: {
      type: {
        "Valkyrie": 2,
        "Golem": 3,
        "Skeleton": 4,
        "SkeletonMage": 5,
        "Revive": 6,
        "Mercenary": 7,
        "Dopplezon": 8,
        "Raven": 10,
        "SpiritWolf": 11,
        "Fenris": 12,
        "DireWolf": 12,
        "Totem": 13,
        "Spirit": 13,
        "Vine": 14,
        "Grizzly": 15,
        "ShadowWarrior": 16,
        "Shadow": 16,
        "AssassinTrap": 17,
        "Hydra": 19,
      },

      mode: {
        Death: 0,
        Standing: 1,
        Walking: 2,
        GettingHit: 3,
        Attacking1: 4,
        Attacking2: 5,
        Blocking: 6,
        CastingSkill: 7,
        UsingSkill1: 8,
        UsingSkill2: 9,
        UsingSkill3: 10,
        UsingSkill4: 11,
        Dead: 12,
        KnockedBack: 13,
        Spawning: 14,
        Running: 15
      },

      ClayGolem: 289,
      Dopplezon: 356,
      Valkyrie: 357,
      FireGolem: 292,
      IronGolem: 291,
      NecroMage: 364,
      NecroSkeleton: 363,
      Poppy: 425,
      Wolverine: 423,
    },

    mercs: {
      mode: {
        Death: 0,
        Standing: 1,
        Walking: 2,
        GettingHit: 3,
        Attacking1: 4,
        Attacking2: 5,
        Blocking: 6,
        CastingSkill: 7,
        UsingSkill1: 8,
        UsingSkill2: 9,
        UsingSkill3: 10,
        UsingSkill4: 11,
        Dead: 12,
        KnockedBack: 13,
        Spawning: 14,
        Running: 15
      },

      Rogue: 271,
      Guard: 338,
      IronWolf: 359,
      A5Barb: 561,
    },

    missiles: {
      DiabloLightning: 172,
      FissureCrack1: 462,
      FissureCrack2: 463,
    },

    storage: {
      Equipped: 1,
      Belt: 2,
      Inventory: 3,
      TradeWindow: 5,
      Cube: 6,
      Stash: 7,
    },

    node: {
      NotOnPlayer: 0,
      Storage: 1,
      Belt: 2,
      Equipped: 3,
      Cursor: 4,
    },

    // Same apply's for merc with less things available
    body: {
      None: 0,
      Head: 1,
      Neck: 2,
      Torso: 3,
      Armor: 3,
      RightArm: 4,
      LeftArm: 5,
      RingRight: 6,
      RingLeft: 7,
      Belt: 8,
      Feet: 9,
      Gloves: 10,
      RightArmSecondary: 11,
      LeftArmSecondary: 12
    },

    items: {
      cost: {
        ToBuy: 0,
        ToSell: 1,
        ToRepair: 2,
      },
      flags: {
        Equipped: 0x00000001,
        InSocket: 0x00000008,
        Identified: 0x00000010,
        OnActiveWeaponSlot: 0x00000040,
        OnSwapWeaponSlot: 0x00000080,
        Broken: 0x00000100,
        FullRejuv: 0x00000400,
        Socketed: 0x00000800,
        InTradeGamble: 0x00002000,
        NotInSocket: 0x00004000,
        Ear: 0x00010000,
        StartingItem: 0x00020000,
        RuneQuestPotion: 0x00200000,
        Ethereal: 0x00400000,
        IsAnItem: 0x00800000,
        Personalized: 0x01000000,
        Runeword: 0x04000000,
      },
      mode: {
        inStorage: 0, //Item inven stash cube store = Item inven stash cube store
        Equipped: 1, // Item equipped self or merc
        inBelt: 2, // Item in belt
        onGround: 3, // Item on ground
        onCursor: 4, // Item on cursor
        Dropping: 5, // Item being dropped
        Socketed: 6 // Item socketed in item
      },
      quality: {
        LowQuality: 1,
        Normal: 2,
        Superior: 3,
        Magic: 4,
        Set: 5,
        Rare: 6,
        Unique: 7,
        Crafted: 8,
      },
      class: {
        Normal: 0,
        Exceptional: 1,
        Elite: 2,
      },
      type: {
        Shield: 2,
        Armor: 3,
        Gold: 4,
        BowQuiver: 5,
        CrossbowQuiver: 6,
        PlayerBodyPart: 7,
        Herb: 8,
        Potion: 9,
        Ring: 10,
        Elixir: 11,
        Amulet: 12,
        Charm: 13,
        notused0: 14,
        Boots: 15,
        Gloves: 16,
        notused1: 17,
        Book: 18,
        Belt: 19,
        Gem: 20,
        Torch: 21,
        Scroll: 22,
        notused2: 23,
        Scepter: 24,
        Wand: 25,
        Staff: 26,
        Bow: 27,
        Axe: 28,
        Club: 29,
        Sword: 30,
        Hammer: 31,
        Knife: 32,
        Spear: 33,
        Polearm: 34,
        Crossbow: 35,
        Mace: 36,
        Helm: 37,
        MissilePotion: 38,
        Quest: 39,
        Bodypart: 40,
        Key: 41,
        ThrowingKnife: 42,
        ThrowingAxe: 43,
        Javelin: 44,
        Weapon: 45,
        MeleeWeapon: 46,
        MissileWeapon: 47,
        ThrownWeapon: 48,
        ComboWeapon: 49,
        AnyArmor: 50,
        AnyShield: 51,
        Miscellaneous: 52,
        SocketFiller: 53,
        Secondhand: 54,
        StavesandRods: 55,
        Missile: 56,
        Blunt: 57,
        Jewel: 58,
        ClassSpecific: 59,
        AmazonItem: 60,
        BarbarianItem: 61,
        NecromancerItem: 62,
        PaladinItem: 63,
        SorceressItem: 64,
        AssassinItem: 65,
        DruidItem: 66,
        HandtoHand: 67,
        Orb: 68,
        VoodooHeads: 69,
        AuricShields: 70,
        PrimalHelm: 71,
        Pelt: 72,
        Cloak: 73,
        Rune: 74,
        Circlet: 75,
        HealingPotion: 76,
        ManaPotion: 77,
        RejuvPotion: 78,
        StaminaPotion: 79,
        AntidotePotion: 80,
        ThawingPotion: 81,
        SmallCharm: 82,
        LargeCharm: 83,
        GrandCharm: 84,
        AmazonBow: 85,
        AmazonSpear: 86,
        AmazonJavelin: 87,
        AssassinClaw: 88,
        MagicBowQuiv: 89,
        MagicxBowQuiv: 90,
        ChippedGem: 91,
        FlawedGem: 92,
        StandardGem: 93,
        FlawlessGem: 94,
        PerfectgGem: 95,
        Amethyst: 96,
        Diamond: 97,
        Emerald: 98,
        Ruby: 99,
        Sapphire: 100,
        Topaz: 101,
        Skull: 102,
      },
      
      // Weapons
      "HandAxe": 0,
      "Axe": 1,
      "DoubleAxe": 2,
      "MilitaryPick": 3,
      "WarAxe": 4,
      "LargeAxe": 5,
      "BroadAxe": 6,
      "BattleAxe": 7,
      "GreatAxe": 8,
      "GiantAxe": 9,
      "Wand": 10,
      "YewWand": 11,
      "BoneWand": 12,
      "GrimWand": 13,
      "Club": 14,
      "Scepter": 15,
      "GrandScepter": 16,
      "WarScepter": 17,
      "SpikedClub": 18,
      "Mace": 19,
      "MorningStar": 20,
      "Flail": 21,
      "WarHammer": 22,
      "Maul": 23,
      "GreatMaul": 24,
      "ShortSword": 25,
      "Scimitar": 26,
      "Sabre": 27,
      "Falchion": 28,
      "CrystalSword": 29,
      "BroadSword": 30,
      "LongSword": 31,
      "WarSword": 32,
      "Two_HandedSword": 33,
      "Claymore": 34,
      "GiantSword": 35,
      "BastardSword": 36,
      "Flamberge": 37,
      "GreatSword": 38,
      "Dagger": 39,
      "Dirk": 40,
      "Kris": 41,
      "Blade": 42,
      "ThrowingKnife": 43,
      "ThrowingAxe": 44,
      "BalancedKnife": 45,
      "BalancedAxe": 46,
      "Javelin": 47,
      "Pilum": 48,
      "ShortSpear": 49,
      "Glaive": 50,
      "ThrowingSpear": 51,
      "Spear": 52,
      "Trident": 53,
      "Brandistock": 54,
      "Spetum": 55,
      "Pike": 56,
      "Bardiche": 57,
      "Voulge": 58,
      "Scythe": 59,
      "Poleaxe": 60,
      "Halberd": 61,
      "WarScythe": 62,
      "ShortStaff": 63,
      "LongStaff": 64,
      "GnarledStaff": 65,
      "BattleStaff": 66,
      "WarStaff": 67,
      "ShortBow": 68,
      "HuntersBow": 69,
      "LongBow": 70,
      "CompositeBow": 71,
      "ShortBattleBow": 72,
      "LongBattleBow": 73,
      "ShortWarBow": 74,
      "LongWarBow": 75,
      "LightCrossbow": 76,
      "Crossbow": 77,
      "HeavyCrossbow": 78,
      "RepeatingCrossbow": 79,
      "Hatchet": 93,
      "Cleaver": 94,
      "TwinAxe": 95,
      "Crowbill": 96,
      "Naga": 97,
      "MilitaryAxe": 98,
      "BeardedAxe": 99,
      "Tabar": 100,
      "GothicAxe": 101,
      "AncientAxe": 102,
      "BurntWand": 103,
      "PetrifiedWand": 104,
      "TombWand": 105,
      "GraveWand": 106,
      "Cudgel": 107,
      "RuneScepter": 108,
      "HolyWaterSprinkler": 109,
      "DivineScepter": 110,
      "BarbedClub": 111,
      "FlangedMace": 112,
      "JaggedStar": 113,
      "Knout": 114,
      "BattleHammer": 115,
      "WarClub": 116,
      "MarteldeFer": 117,
      "Gladius": 118,
      "Cutlass": 119,
      "Shamshir": 120,
      "Tulwar": 121,
      "DimensionalBlade": 122,
      "BattleSword": 123,
      "RuneSword": 124,
      "AncientSword": 125,
      "Espandon": 126,
      "DacianFalx": 127,
      "TuskSword": 128,
      "GothicSword": 129,
      "Zweihander": 130,
      "ExecutionerSword": 131,
      "Poignard": 132,
      "Rondel": 133,
      "Cinquedeas": 134,
      "Stiletto": 135,
      "BattleDart": 136,
      "Francisca": 137,
      "WarDart": 138,
      "Hurlbat": 139,
      "WarJavelin": 140,
      "GreatPilum": 141,
      "Simbilan": 142,
      "Spiculum": 143,
      "Harpoon": 144,
      "WarSpear": 145,
      "Fuscina": 146,
      "WarFork": 147,
      "Yari": 148,
      "Lance": 149,
      "LochaberAxe": 150,
      "Bill": 151,
      "BattleScythe": 152,
      "Partizan": 153,
      "Bec_de_Corbin": 154,
      "GrimScythe": 155,
      "JoStaff": 156,
      "Quarterstaff": 157,
      "CedarStaff": 158,
      "GothicStaff": 159,
      "RuneStaff": 160,
      "EdgeBow": 161,
      "RazorBow": 162,
      "CedarBow": 163,
      "DoubleBow": 164,
      "ShortSiegeBow": 165,
      "LargeSiegeBow": 166,
      "RuneBow": 167,
      "GothicBow": 168,
      "Arbalest": 169,
      "SiegeCrossbow": 170,
      "Ballista": 171,
      "Chu_Ko_Nu": 172,
      "Katar": 175,
      "WristBlade": 176,
      "HatchetHands": 177,
      "Cestus": 178,
      "Claws": 179,
      "BladeTalons": 180,
      "ScissorsKatar": 181,
      "Quhab": 182,
      "WristSpike": 183,
      "Fascia": 184,
      "HandScythe": 185,
      "GreaterClaws": 186,
      "GreaterTalons": 187,
      "ScissorsQuhab": 188,
      "Suwayyah": 189,
      "WristSword": 190,
      "WarFist": 191,
      "BattleCestus": 192,
      "FeralClaws": 193,
      "RunicTalons": 194,
      "ScissorsSuwayyah": 195,
      "Tomahawk": 196,
      "SmallCrescent": 197,
      "EttinAxe": 198,
      "WarSpike": 199,
      "BerserkerAxe": 200,
      "FeralAxe": 201,
      "Silver_edgedAxe": 202,
      "Decapitator": 203,
      "ChampionAxe": 204,
      "GloriousAxe": 205,
      "PolishedWand": 206,
      "GhostWand": 207,
      "LichWand": 208,
      "UnearthedWand": 209,
      "Truncheon": 210,
      "MightyScepter": 211,
      "SeraphRod": 212,
      "Caduceus": 213,
      "TyrantClub": 214,
      "ReinforcedMace": 215,
      "DevilStar": 216,
      "Scourge": 217,
      "LegendaryMallet": 218,
      "OgreMaul": 219,
      "ThunderMaul": 220,
      "Falcata": 221,
      "Ataghan": 222,
      "ElegantBlade": 223,
      "HydraEdge": 224,
      "PhaseBlade": 225,
      "ConquestSword": 226,
      "CrypticSword": 227,
      "MythicalSword": 228,
      "LegendSword": 229,
      "HighlandBlade": 230,
      "BalrogBlade": 231,
      "ChampionSword": 232,
      "ColossusSword": 233,
      "ColossusBlade": 234,
      "BoneKnife": 235,
      "MithrilPoint": 236,
      "FangedKnife": 237,
      "LegendSpike": 238,
      "FlyingKnife": 239,
      "FlyingAxe": 240,
      "WingedKnife": 241,
      "WingedAxe": 242,
      "HyperionJavelin": 243,
      "StygianPilum": 244,
      "BalrogSpear": 245,
      "GhostGlaive": 246,
      "WingedHarpoon": 247,
      "HyperionSpear": 248,
      "StygianPike": 249,
      "Mancatcher": 250,
      "GhostSpear": 251,
      "WarPike": 252,
      "OgreAxe": 253,
      "ColossusVoulge": 254,
      "Thresher": 255,
      "CrypticAxe": 256,
      "GreatPoleaxe": 257,
      "GiantThresher": 258,
      "WalkingStick": 259,
      "Stalagmite": 260,
      "ElderStaff": 261,
      "Shillelagh": 262,
      "ArchonStaff": 263,
      "SpiderBow": 264,
      "BladeBow": 265,
      "ShadowBow": 266,
      "GreatBow": 267,
      "DiamondBow": 268,
      "CrusaderBow": 269,
      "WardBow": 270,
      "HydraBow": 271,
      "PelletBow": 272,
      "GorgonCrossbow": 273,
      "ColossusCrossbow": 274,
      "DemonCrossbow": 275,
      "EagleOrb": 276,
      "SacredGlobe": 277,
      "SmokedSphere": 278,
      "ClaspedOrb": 279,
      "JaredsStone": 280,
      "StagBow": 281,
      "ReflexBow": 282,
      "MaidenSpear": 283,
      "MaidenPike": 284,
      "MaidenJavelin": 285,
      "GlowingOrb": 286,
      "CrystallineGlobe": 287,
      "CloudySphere": 288,
      "SparklingBall": 289,
      "SwirlingCrystal": 290,
      "AshwoodBow": 291,
      "CeremonialBow": 292,
      "CeremonialSpear": 293,
      "CeremonialPike": 294,
      "CeremonialJavelin": 295,
      "HeavenlyStone": 296,
      "EldritchOrb": 297,
      "DemonHeart": 298,
      "VortexOrb": 299,
      "DimensionalShard": 300,
      "MatriarchalBow": 301,
      "GrandMatronBow": 302,
      "MatriarchalSpear": 303,
      "MatriarchalPike": 304,
      "MatriarchalJavelin": 305,
      "Cap": 306,
      "SkullCap": 307,
      "Helm": 308,
      "FullHelm": 309,
      "GreatHelm": 310,
      "Crown": 311,
      "Mask": 312,
      "QuiltedArmor": 313,
      "LeatherArmor": 314,
      "HardLeatherArmor": 315,
      "StuddedLeather": 316,
      "RingMail": 317,
      "ScaleMail": 318,
      "ChainMail": 319,
      "BreastPlate": 320,
      "SplintMail": 321,
      "PlateMail": 322,
      "FieldPlate": 323,
      "GothicPlate": 324,
      "FullPlateMail": 325,
      "AncientArmor": 326,
      "LightPlate": 327,
      "Buckler": 328,
      "SmallShield": 329,
      "LargeShield": 330,
      "KiteShield": 331,
      "TowerShield": 332,
      "GothicShield": 333,
      "LeatherGloves": 334,
      "HeavyGloves": 335,
      "ChainGloves": 336,
      "LightGauntlets": 337,
      "Gauntlets": 338,
      "Boots": 339,
      "HeavyBoots": 340,
      "ChainBoots": 341,
      "LightPlatedBoots": 342,
      "Greaves": 343,
      "Sash": 344,
      "LightBelt": 345,
      "Belt": 346,
      "HeavyBelt": 347,
      "PlatedBelt": 348,
      "BoneHelm": 349,
      "BoneShield": 350,
      "SpikedShield": 351,
      "WarHat": 352,
      "Sallet": 353,
      "Casque": 354,
      "Basinet": 355,
      "WingedHelm": 356,
      "GrandCrown": 357,
      "DeathMask": 358,
      "GhostArmor": 359,
      "SerpentskinArmor": 360,
      "DemonhideArmor": 361,
      "TrellisedArmor": 362,
      "LinkedMail": 363,
      "TigulatedMail": 364,
      "MeshArmor": 365,
      "Cuirass": 366,
      "RussetArmor": 367,
      "TemplarCoat": 368,
      "SharktoothArmor": 369,
      "EmbossedPlate": 370,
      "ChaosArmor": 371,
      "OrnatePlate": 372,
      "MagePlate": 373,
      "Defender": 374,
      "RoundShield": 375,
      "Scutum": 376,
      "DragonShield": 377,
      "Pavise": 378,
      "AncientShield": 379,
      "DemonhideGloves": 380,
      "SharkskinGloves": 381,
      "HeavyBracers": 382,
      "BattleGauntlets": 383,
      "WarGauntlets": 384,
      "DemonhideBoots": 385,
      "SharkskinBoots": 386,
      "MeshBoots": 387,
      "BattleBoots": 388,
      "WarBoots": 389,
      "DemonhideSash": 390,
      "SharkskinBelt": 391,
      "MeshBelt": 392,
      "BattleBelt": 393,
      "WarBelt": 394,
      "GrimHelm": 395,
      "GrimShield": 396,
      "BarbedShield": 397,
      "WolfHead": 398,
      "HawkHelm": 399,
      "Antlers": 400,
      "FalconMask": 401,
      "SpiritMask": 402,
      "JawboneCap": 403,
      "FangedHelm": 404,
      "HornedHelm": 405,
      "AssaultHelmet": 406,
      "AvengerGuard": 407,
      "Targe": 408,
      "Rondache": 409,
      "HeraldicShield": 410,
      "AerinShield": 411,
      "CrownShield": 412,
      "PreservedHead": 413,
      "ZombieHead": 414,
      "UnravellerHead": 415,
      "GargoyleHead": 416,
      "DemonHead": 417,
      "Circlet": 418,
      "Coronet": 419,
      "Tiara": 420,
      "Diadem": 421,
      "Shako": 422,
      "Hydraskull": 423,
      "Armet": 424,
      "GiantConch": 425,
      "SpiredHelm": 426,
      "Corona": 427,
      "Demonhead": 428,
      "DuskShroud": 429,
      "Wyrmhide": 430,
      "ScarabHusk": 431,
      "WireFleece": 432,
      "DiamondMail": 433,
      "LoricatedMail": 434,
      "Boneweave": 435,
      "GreatHauberk": 436,
      "BalrogSkin": 437,
      "HellforgePlate": 438,
      "KrakenShell": 439,
      "LacqueredPlate": 440,
      "ShadowPlate": 441,
      "SacredArmor": 442,
      "ArchonPlate": 443,
      "Heater": 444,
      "Luna": 445,
      "Hyperion": 446,
      "Monarch": 447,
      "Aegis": 448,
      "Ward": 449,
      "BrambleMitts": 450,
      "VampireboneGloves": 451,
      "Vambraces": 452,
      "CrusaderGauntlets": 453,
      "OgreGauntlets": 454,
      "WyrmhideBoots": 455,
      "ScarabshellBoots": 456,
      "BoneweaveBoots": 457,
      "MirroredBoots": 458,
      "MyrmidonGreaves": 459,
      "SpiderwebSash": 460,
      "VampirefangBelt": 461,
      "MithrilCoil": 462,
      "TrollBelt": 463,
      "ColossusGirdle": 464,
      "BoneVisage": 465,
      "TrollNest": 466,
      "BladeBarrier": 467,
      "AlphaHelm": 468,
      "GriffonHeaddress": 469,
      "HuntersGuise": 470,
      "SacredFeathers": 471,
      "TotemicMask": 472,
      "JawboneVisor": 473,
      "LionHelm": 474,
      "RageMask": 475,
      "SavageHelmet": 476,
      "SlayerGuard": 477,
      "AkaranTarge": 478,
      "AkaranRondache": 479,
      "ProtectorShield": 480,
      "GildedShield": 481,
      "RoyalShield": 482,
      "MummifiedTrophy": 483,
      "FetishTrophy": 484,
      "SextonTrophy": 485,
      "CantorTrophy": 486,
      "HierophantTrophy": 487,
      "BloodSpirit": 488,
      "SunSpirit": 489,
      "EarthSpirit": 490,
      "SkySpirit": 491,
      "DreamSpirit": 492,
      "CarnageHelm": 493,
      "FuryVisor": 494,
      "DestroyerHelm": 495,
      "ConquerorCrown": 496,
      "GuardianCrown": 497,
      "SacredTarge": 498,
      "SacredRondache": 499,
      "KurastShield": 500,
      "ZakarumShield": 501,
      "VortexShield": 502,
      "MinionSkull": 503,
      "HellspawnSkull": 504,
      "OverseerSkull": 505,
      "SuccubusSkull": 506,
      "BloodlordSkull": 507,
      "Amulet": 520,
      "Ring": 522,
      "Arrows": 526,
      "Bolts": 528,
      "Jewel": 643,

      // Misc?
      "Elixir": 508,
      "Torch": 527,
      "Heart": 531,
      "Brain": 532,
      "Jawbone": 533,
      "Eye": 534,
      "Horn": 535,
      "Tail": 536,
      "Flag": 537,
      "Fang": 538,
      "Quill": 539,
      "Soul": 540,
      "Scalp": 541,
      "Spleen": 542,
      "Key": 543,
      "Ear": 556,
      "Herb": 602,
      "anevilforce": 609,
      // Potions, tomes/scrolls, gold
      "TomeofTownPortal": 518,
      "TomeofIdentify": 519,
      "ScrollofTownPortal": 529,
      "ScrollofIdentify": 530,
      "RancidGasPotion": 80,
      "OilPotion": 81,
      "ChokingGasPotion": 82,
      "ExplodingPotion": 83,
      "StranglingGasPotion": 84,
      "FulminatingPotion": 85,
      "StaminaPotion": 513,
      "AntidotePotion": 514,
      "RejuvenationPotion": 515,
      "FullRejuvenationPotion": 516,
      "ThawingPotion": 517,
      "MinorHealingPotion": 587,
      "LightHealingPotion": 588,
      "HealingPotion": 589,
      "GreaterHealingPotion": 590,
      "SuperHealingPotion": 591,
      "MinorManaPotion": 592,
      "LightManaPotion": 593,
      "ManaPotion": 594,
      "GreaterManaPotion": 595,
      "SuperManaPotion": 596,
      "Gold": 523,
      // Charms
      "SmallCharm": 603,
      "LargeCharm": 604,
      "GrandCharm": 605,

      quest: {
        // Act 1
        WirtsLeg: 88,
        HoradricMalus: 89,
        ScrollofInifuss: 524,
        KeytotheCairnStones: 525,
        // Act 2
        FinishedStaff: 91,
        "HoradricStaff": 91,
        IncompleteStaff: 92,
        "ShaftoftheHoradricStaff": 92,
        ViperAmulet: 521,
        "TopoftheHoradricStaff": 521,
        Cube: 549,
        BookofSkill: 552,
        // Act 3
        "DecoyGidbinn": 86,
        "TheGidbinn": 87,
        KhalimsFlail: 173,
        KhalimsWill: 174,
        PotofLife: 545,
        "AJadeFigurine": 546,
        JadeFigurine: 546,
        TheGoldenBird: 547,
        LamEsensTome: 548,
        KhalimsEye: 553,
        KhalimsHeart: 554,
        KhalimsBrain: 555,
        // Act 4
        HellForgeHammer: 90,
        Soulstone: 551,
        "MephistosSoulstone": 551,
        // Act 5
        MalahsPotion: 644,
        "ScrollofKnowledge": 645,
        ScrollofResistance: 646,
        // Pandemonium Event
        KeyofTerror: 647,
        KeyofHate: 648,
        KeyofDestruction: 649,
        DiablosHorn: 650,
        BaalsEye: 651,
        MephistosBrain: 652,
        StandardofHeroes: 658,
        // Essences/Token
        TokenofAbsolution: 653,
        TwistedEssenceofSuffering: 654,
        ChargedEssenceofHatred: 655,
        BurningEssenceofTerror: 656,
        FesteringEssenceofDestruction: 657,
        // Misc
        "TheBlackTowerKey": 544,
      },
      runes: {
        El: 610,
        Eld: 611,
        Tir: 612,
        Nef: 613,
        Eth: 614,
        Ith: 615,
        Tal: 616,
        Ral: 617,
        Ort: 618,
        Thul: 619,
        Amn: 620,
        Sol: 621,
        Shael: 622,
        Dol: 623,
        Hel: 624,
        Io: 625,
        Lum: 626,
        Ko: 627,
        Fal: 628,
        Lem: 629,
        Pul: 630,
        Um: 631,
        Mal: 632,
        Ist: 633,
        Gul: 634,
        Vex: 635,
        Ohm: 636,
        Lo: 637,
        Sur: 638,
        Ber: 639,
        Jah: 640,
        Cham: 641,
        Zod: 642,
      },
      gems: {
        Perfect: {
          Amethyst: 561,
          Topaz: 566,
          Sapphire: 571,
          Emerald: 576,
          Ruby: 581,
          Diamond: 586,
          Skull: 601,
        },
        Flawless: {
          Amethyst: 560,
          Topaz: 565,
          Sapphire: 570,
          Emerald: 575,
          Ruby: 580,
          Diamond: 585,
          Skull: 600,
        },
        Normal: {
          Amethyst: 559,
          Topaz: 564,
          Sapphire: 569,
          Emerald: 574,
          Ruby: 579,
          Diamond: 584,
          Skull: 599,
        },
        Flawed: {
          Amethyst: 558,
          Topaz: 563,
          Sapphire: 568,
          Emerald: 573,
          Ruby: 578,
          Diamond: 583,
          Skull: 598,
        },
        Chipped: {
          Amethyst: 557,
          Topaz: 562,
          Sapphire: 567,
          Emerald: 572,
          Ruby: 577,
          Diamond: 582,
          Skull: 597,
        },
      },
    },

    // locale strings
    locale: {
      monsters: {
        // bosses
        Andariel: 3021,
        Duriel: 3054,
        Mephisto: 3062,
        Diablo: 3060,
        Baal: 3061,
        // Mini bosses
        BloodRaven: 3111,
        TreeheadWoodFist: 2873,
        TheCountess: 2875,
        TheSmith: 2889,
        Radament: 2879,
        TheSummoner: 3099,
        HephastoTheArmorer: 1067,
        Izual: 1014,
        ShenktheOverseer: 22435,
        // Uniques
        Corpsefire: 3319,
        TheCowKing: 2850,
        GrandVizierofChaos: 2851,
        LordDeSeis: 2852,
        InfectorofSouls: 2853,
        RiftwraiththeCannibal: 2854,
        Taintbreeder: 2855,
        TheTormentor: 2856,
        Darkwing: 2857,
        MafferDragonhand: 2858,
        WyandVoidbringer: 2859,
        ToorcIcefist: 2860,
        BremmSparkfist: 2861,
        GelebFlamefinger: 2862,
        IsmailVilehand: 2863,
        IcehawkRiftwing: 2864,
        BattlemaidSarina: 2865,
        Stormtree: 2866,
        WitchDoctorEndugu: 2867,
        SszarkTheBurning: 2868,
        Bishibosh: 2869,
        Bonebreaker: 2870,
        Coldcrow: 2871,
        Rakanishu: 2872,
        Griswold: 2874,
        PitspawnFouldog: 2876,
        FlamespiketheCrawler: 2877,
        BoneAsh: 2878,
        BloodwitchtheWild: 2880,
        Fangskin: 2881,
        Beetleburst: 2882,
        CreepingFeature: 2883,
        ColdwormtheBurrower: 2884,
        FireEye: 2885,
        DarkElder: 2886,
        AncientKaatheSoulless: 2888,
        SharpToothSayer: 22493,
        SnapchipShatter: 22496,
        Pindleskin: 22497,
        ThreshSocket: 22498,
        EyebacktheUnleashed: 22499,
        EldritchtheRectifier: 22500,
        DacFarren: 22501,
        BonesawBreaker: 22502,
        Frozenstein: 22504,
        Rogue: 2897,
        StygianDoll: 2898,
        SoulKiller: 2899,
        Flayer: 2900,
        Fetish: 2901,
        RatMan: 2902,
        UndeadStygianDoll: 2903,
        UndeadSoulKiller: 2904,
        UndeadFlayer: 2905,
        UndeadFetish: 2906,
        UndeadRatMan: 2907,
        DarkFamiliar: 2908,
        BloodDiver: 2909,
        Gloombat: 2910,
        DesertWing: 2911,
        TheBanished: 2912,
        BloodLord: 2913,
        DarkLord: 2914,
        NightLord: 2915,
        GhoulLord: 2916,
        Spikefist: 2917,
        Thrasher: 2918,
        BrambleHulk: 2919,
        ThornedHulk: 2920,
        SpiderMagus: 2921,
        FlameSpider: 2922,
        PoisonSpinner: 2923,
        SandFisher: 2924,
        Arach: 2925,
        BloodWing: 2926,
        BloodHook: 2927,
        Feeder: 2928,
        Sucker: 2929,
        WingedNightmare: 2930,
        HellBuzzard: 2931,
        UndeadScavenger: 2932,
        CarrionBird: 2933,
        Unraveler: 2934,
        Guardian: 2935,
        HollowOne: 2936,
        HoradrimAncient: 2937,
        BoneScarab: 2938,
        SteelScarab: 2939,
        Scarab: 2940,
        DeathBeetle: 2941,
        DungSoldier: 2942,
        HellSwarm: 2943,
        PlagueBugs: 2944,
        BlackLocusts: 2945,
        Itchies: 2946,
        HellCat: 2947,
        NightTiger: 2948,
        SaberCat: 2949,
        Huntress: 2950,
        CliffLurker: 2951,
        TreeLurker: 2952,
        CaveLeaper: 2953,
        TombCreeper: 2954,
        SandLeaper: 2955,
        TombViper: 2956,
        PitViper: 2957,
        Salamander: 2958,
        ClawViper: 2959,
        SerpentMagus: 2960,
        BloodMaggot: 2961,
        GiantLamprey: 2962,
        Devourer: 2963,
        RockWorm: 2964,
        SandMaggot: 2965,
        BushBarb: 2966,
        RazorSpine: 2967,
        ThornBeast: 2968,
        SpikeFiend: 2969,
        QuillRat: 2970,
        HellClan: 2971,
        MoonClan: 2972,
        NightClan: 2973,
        DeathClan: 2974,
        BloodClan: 2975,
        TempleGuard: 2976,
        DoomApe: 2977,
        JungleHunter: 2978,
        RockDweller: 2979,
        DuneBeast: 2980,
        FleshHunter: 2981,
        BlackRogue: 2982,
        DarkStalker: 2983,
        VileHunter: 2984,
        DarkHunter: 2985,
        DarkShape: 2986,
        Apparition: 2987,
        Specter: 2988,
        Wraith: 2989,
        Ghost: 2990,
        Assailant: 2991,
        Infidel: 2992,
        Invader: 2993,
        Marauder: 2994,
        SandRaider: 2995,
        GargantuanBeast: 2996,
        WailingBeast: 2997,
        Yeti: 2998,
        Crusher: 2999,
        Brute: 3000,
        CloudStalker: 3001,
        BlackVulture: 3002,
        BlackRaptor: 3003,
        BloodHawk: 3004,
        FoulCrow: 3005,
        PlagueBearer: 3006,
        Ghoul: 3007,
        DrownedCarcass: 3008,
        HungryDead: 3009,
        Zombie: 3010,
        Horror: 3012,
        Returned: 3013,
        BurningDead: 3014,
        BoneWarrior: 3015,
        Damned: 3016,
        Disfigured: 3017,
        Misshapen: 3018,
        Tainted: 3019,
        Afflicted: 3020,
        Camel: 3033,
        Cadaver: 3034,
        PreservedDead: 3035,
        Embalmed: 3036,
        DriedCorpse: 3037,
        Decayed: 3038,
        Urdar: 3039,
        Mauler: 3040,
        Gorebelly: 3041,
        Blunderbore: 3042,
        BloodMaggotYoung: 3043,
        GiantLampreyYoung: 3044,
        DevourerYoung: 3045,
        RockWormYoung: 3046,
        SandMaggotYoung: 3047,
        BloodMaggotEgg: 3048,
        GiantLampreyEgg: 3049,
        DevourerEgg: 3050,
        RockWormEgg: 3051,
        SandMaggotEgg: 3052,
        Maggot: 3053,
        BloodHawkNest: 3055,
        FlyingScimitar: 3056,
        CloudStalkerNest: 3057,
        BlackRaptorNest: 3058,
        FoulCrowNest: 3059,
        Cantor: 3063,
        Heirophant: 3064,
        Sexton: 3065,
        Zealot: 3066,
        Faithful: 3067,
        Zakarumite: 3068,
        BlackSoul: 3069,
        BurningSoul: 3070,
        SwampGhost: 3071,
        Gloam: 3072,
        WarpedShaman: 3073,
        DarkShaman: 3074,
        DevilkinShaman: 3075,
        CarverShaman: 3076,
        FallenShaman: 3077,
        WarpedOne: 3078,
        DarkOne: 3079,
        Devilkin: 3080,
        Carver: 3081,
        Fallen: 3082,
        ReturnedArcher: 3083,
        HorrorArcher: 3084,
        BurningDeadArcher: 3085,
        BoneArcher: 3086,
        CorpseArcher: 3087,
        SkeletonArcher: 3088,
        FleshLancer: 3089,
        BlackLancer: 3090,
        DarkLancer: 3091,
        VileLancer: 3092,
        DarkSpearwoman: 3093,
        FleshArcher: 3094,
        BlackArcher: 3095,
        DarkRanger: 3096,
        VileArcher: 3097,
        DarkArcher: 3098,
        StygianDollShaman: 3100,
        SoulKillerShaman: 3101,
        FlayerShaman: 3102,
        FetishShaman: 3103,
        RatManShaman: 3104,
        HorrorMage: 3105,
        BurningDeadMage: 3106,
        BoneMage: 3107,
        CorpseMage: 3108,
        ReturnedMage: 3109,
        GargoyleTrap: 3110,
        NightMarauder: 3121,
        FireGolem: 3122,
        IronGolem: 3123,
        BloodGolem: 3124,
        ClayGolem: 3125,
        BloodMaggotQueen: 3126,
        GiantLampreyQueen: 3127,
        DevourerQueen: 3128,
        RockWormQueen: 3129,
        SandMaggotQueen: 3130,
        SlimePrince: 3131,
        BogCreature: 3132,
        SwampDweller: 3133,
        BarbedGiant: 3134,
        RazorBeast: 3135,
        ThornBrute: 3136,
        SpikeGiant: 3137,
        QuillBear: 3138,
        CouncilMember: 3139,
        DarkWanderer: 3141,
        HellSlinger: 3142,
        NightSlinger: 3143,
        SpearCat: 3144,
        Slinger: 3145,
        FireTower: 3146,
        LightningSpire: 3147,
        PitLord: 3148,
        Balrog: 3149,
        VenomLord: 3150,
        IronWolf: 3151,
        InvisoSpawner: 3152,
        OblivionKnight: 3153,
        Mage: 3154,
        AbyssKnight: 3155,
        FighterMage: 3156,
        DoomKnight: 3157,
        Fighter: 3158,
        MawFiend: 3159,
        CorpseSpitter: 3160,
        Corpulent: 3161,
        StormCaster: 3162,
        Strangler: 3163,
        DoomCaster: 3164,
        GrotesqueWyrm: 3165,
        StygianDog: 3166,
        FleshBeast: 3167,
        Grotesque: 3168,
        StygianHag: 3169,
        FleshSpawner: 3170,
        RogueScout: 3171,
        BloodWingNest: 3172,
        BloodHookNest: 3173,
        FeederNest: 3174,
        SuckerNest: 3175,
        Hydra: 3325,
      },
      npcs: {
        Asheara: 1008,
        Hratli: 1009,
        Alkor: 1010,
        Ormus: 1011,
        Natalya: 1012,
        Tyrael: 1013,
        Izual1: 1014,
        Izual2: 1015,
        Jamella: 1016,
        Halbu: 1017,
        Hadriel: 1018,
        Hazade: 1019,
        Alhizeer: 1020,
        Azrael: 1021,
        Ahsab: 1022,
        Chalan: 1023,
        Haseen: 1024,
        Razan: 1025,
        Emilio: 1026,
        Pratham: 1027,
        Fazel: 1028,
        Jemali: 1029,
        Kasim: 1030,
        Gulzar: 1031,
        Mizan: 1032,
        Leharas: 1033,
        Durga: 1034,
        Neeraj: 1035,
        Ilzan: 1036,
        Zanarhi: 1037,
        Waheed: 1038,
        Vikhyat: 1039,
        Jelani: 1040,
        Barani: 1041,
        Jabari: 1042,
        Devak: 1043,
        Raldin: 1044,
        Telash: 1045,
        Ajheed: 1046,
        Narphet: 1047,
        Khaleel: 1048,
        Phaet: 1049,
        Geshef: 1050,
        Vanji: 1051,
        Haphet: 1052,
        Thadar: 1053,
        Yatiraj: 1054,
        Rhadge: 1055,
        Yashied: 1056,
        Lharhad: 1057,
        Flux: 1058,
        Scorch: 1059,
        //Natalya: 3022, both 1012 and 3022 return Natalya?
        DeckardCain: 2890,
        Gheed: 2891,
        Akara: 2892,
        Kashya: 2893,
        Charsi: 2894,
        Warriv: 2895,
        Drognan: 3023,
        Atma: 3024,
        Fara: 3025,
        Lysander: 3026,
        Jerhyn: 3028,
        Geglash: 3029,
        Elzix: 3030,
        Greiz: 3031,
        Flavie: 3112,
        Kaelan: 3113,
        Meshif: 3114,
        Larzuk: 22476,
        Anya: 22477,
        Malah: 22478,
        Nihlathak1: 22479,
        QualKehk: 22480,
        Guard: 22481,
        Combatant: 22482,
        Nihlathak2: 22483,
      },
      items: {
        KhalimsFlail: 1060,
        KhalimsWill1: 1061,
        KhalimsFlail2: 1062,
        KhalimsWill2: 1063,
        KhalimsEye: 1064,
        KhalimsBrain: 1065,
        KhalimsHeart: 1066,
        ScrollofInifuss: 2216,
        KeytotheCairnStones: 2217,
        AJadeFigurine: 2227,
        TheGoldenBird: 2228,
        LamEsensTome1: 2229,
        LamEsensTome2: 2230,
        HoradricCube: 2231,
        HoradricScroll: 2232,
        MephistosSoulstone: 2233,
        Ear: 2235,
        AmuletoftheViper: 2697,
        StaffofKings: 2698,
        HoradricStaff: 2699,

        // Sets
        // Angelic Rainment
        AngelicsSword: 10172,
        AngelicsArmor: 10173,
        AngelicsRing: 10174,
        AngelicsAmulet: 10175,
        // Arcannas Tricks
        ArcannasAmulet: 10180,
        ArcannasStaff: 10181,
        ArcannasHelmet: 10182,
        ArcannasArmor: 10183,
        // Artic Gear
        ArticsBow: 10176,
        ArticsArmor: 10177,
        ArticsBelt: 10178,
        ArticsGloves: 10179,
        // Berserkers Gear
        BerserkersHelmet: 10166,
        BerserkersAxe: 10167,
        BerserkersArmor: 10168,
        // Cathans Traps
        CathansRing: 10147,
        CathansAmulet: 10148,
        CathansHelmet: 10149,
        CathansArmor: 10150,
        CathansStaff: 10151,
        // Civerbs Gear
        CiverbsShield: 10122,
        CiverbsAmulet: 10123,
        CiverbsScepter: 10124,
        // Clegaws Brace
        ClegawsSword: 10128,
        ClegawsShield: 10129,
        ClegawsGloves: 10130,
        // Deaths Disguise
        DeathsGloves: 10169,
        DeathsBelt: 10170,
        DeathsSword: 10171,
        // Hsarus Defense
        HsarusBoots: 10125,
        HsarusShield: 10126,
        HsarusBelt: 10127,
        // Infernal Tools
        InfernalsHelmet: 10163,
        InfernalsWand: 10164,
        InfernalsBelt: 10165,
        // Irathas Finery
        IrathasBelt: 10131,
        IrathasHelmet: 10132,
        IrathasGloves: 10133,
        IrathasAmulet: 10134,
        // Isenharts Armory
        IsenhartsHelmet: 10135,
        IsenhartsArmor: 10136,
        IsenhartsShield: 10137,
        IsenhartsSword: 10138,
        // Milabrega Regalia
        MilabregasArmor: 10143,
        MilabregasHelmet: 10144,
        MilabregasScepter: 10145,
        MilabregasShield: 10146,
        // Sigons
        SigonsHelmet: 10157,
        SigonsArmor: 10158,
        SigonsGloves: 10159,
        SigonsBoots: 10160,
        SigonsBelt: 10161,
        SigonsShield: 10162,
        // Tancreds
        TancredsPick: 10152,
        TancredsArmor: 10153,
        TancredsBoots: 10154,
        TancredsAmulet: 10155,
        TancredsHelmet: 10156,
        // Vidalas
        VidalasAmulet: 10139,
        VidalasArmor: 10140,
        VidalasBoots: 10141,
        VidalasBow: 10142,

        // LoD Sets
        // Aldurs's Legacy
        AldursHelmet: 21697,
        AldursArmor: 21698,
        AldursBoots: 21700,
        AldursMace: 21847,
        // Bul-Kathos's Children
        BulKathosBlade: 21688,
        BulKathoSword: 21689,
        // Cow Kings's Leathers
        CowKingsHelmet: 21723,
        CowKingsArmor: 21724,
        CowKingsBoots: 21725,
        // Disciples
        DisciplesAmulet: 21717,
        DisciplesGloves: 21718,
        DisciplesBoots: 21719,
        DisciplesArmor: 21720,
        DisciplesBelt: 21721,
        // Griswolds's Legacy
        GriswoldsScepter: 21673,
        GriswoldsShield: 21674,
        GriswoldsArmor: 21675,
        GriswoldsHelmet: 21676,
        // Heaven's Brethren
        HeavensMace: 21823,
        HeavensHelmet: 21824,
        HeavensShield: 21825,
        HeavensArmor: 21826,
        // Hwanin's
        HwaninsHelmet: 21712,
        HwaninsPolearm: 21713,
        HwaninsArmor: 21714,
        HwaninsBelt: 21821,
        // IK
        ImmortalKingsMaul: 21840,
        ImmortalKingsBoots: 21841,
        ImmortalKingsGloves: 21842,
        ImmortalKingsBelt: 21843,
        ImmortalKingsArmor: 21844,
        ImmortalKingsHelmet: 21845,
        // M'avina's
        MavinasHelmet: 21702,
        MavinasArmor: 21703,
        MavinasGloves: 21704,
        MavinasBelt: 21705,
        MavinasBow: 21706,
        // Natalya's
        NatalyasHelmet: 21668,
        NatalyasClaw: 21669,
        NatalyasArmor: 21670,
        NatalyasBoots: 21671,
        // Naj's
        NajsStaff: 21640,
        NajsArmor: 21831,
        NajsHelmet: 21832,
        // Orphan's
        OrphansHelmet: 21731,
        OrphansBelt: 21732,
        OrphansGloves: 21733,
        OrphansShield: 21734,
        // Sanders's
        SandersGloves: 21876,
        SandersBoots: 21877,
        SandersHelmet: 21878,
        SandersWand: 21879,
        // Sazabi's
        SazabisSword: 21708,
        SazabisArmor: 21709,
        SazabisHelmet: 21710,
        // Tal
        TalRashasBelt: 21816,
        TalRashasAmulet: 21817,
        TalRashasArmor: 21818,
        TalRashasOrb: 21819,
        TalRashasHelmet: 21820,
        // Trang-Ouls
        TrangOulsHelmet: 21661,
        TrangOulsShield: 21662,
        TrangOulsArmor: 21664,
        TrangOulsGloves: 21665,
        TrangOulsBelt: 21666,

        // Uniques
        // Quest/Misc
        KeyofTerror: 11146,
        KeyofHate: 11147,
        KeyofDestruction: 11148,
        DiablosHorn: 11149,
        BaalsEye: 11150,
        MephistosBrain: 11151,
        StandardofHeroes: 11152,
        HellfireTorch: 11153,
        Annihilus: 21743,

        // Unique Items
        WitchwildString: 10911,
        TitansRevenge: 21735,
        LycandersAim: 21737,
        ArreatsFace: 21744,
        Homunculus: 21755,
        JalalsMane: 21750,
        HeraldofZakarum: 21758,
        BloodRavensCharge: 21508,
        Windforce: 21635,
        Gimmershred: 21637,
        MedusasGaze: 21516,
        Rockstopper: 21519,
        CrownofThieves: 21522,
        BlackhornsFace: 21523,
        TheSpiritShroud: 21524,
        SkinoftheFlayedOne: 21525,
        IronPelt: 21526,
        SpiritForge: 21527,
        CrowCaw: 21528,
        DurielsShell: 21529,
        SkulldersIre: 21530,
        Toothrow: 21531,
        AtmasWail: 21532,
        BlackHades: 21533,
        Corpsemourn: 21534,
        QueHegans: 21535,
        QueHegansWisdom: 21535,
        Mosers: 21536,
        MosersBlessedCircle: 21536,
        Stormchaser: 21537,
        TiamatsRubuke: 21538,
        GerkesSanctuary: 21539,
        RadamentsSphere: 21540,
        Gravepalm: 21541,
        Ghoulhide: 21542,
        Hellmouth: 21543,
        Infernostride: 21544,
        Waterwalk: 21545,
        Silkweave: 21546,
        WarTraveler: 21547,
        Razortail: 21548,
        GloomsTrap: 21549,
        Snowclash: 21550,
        ThundergodsVigor: 21551,
        LidlessWall: 21552,
        LanceGuard: 21553,
        Boneflame: 21555,
        SteelPillar: 21556,
        NightwingsVeil: 21557,
        CrownofAges: 21559,
        AndarielsVisage: 21560,
        Dragonscale: 21562,
        SteelCarapace: 21563,
        RainbowFacet: 21565,
        Ravenlore: 21566,
        Boneshade: 21567,
        Flamebellow: 21570,
        DeathsFathom: 21571,
        Wolfhowl: 21572,
        SpiritWard: 21573,
        KirasGuardian: 21574,
        OrmusRobe: 21575,
        GheedsFortune: 21576,
        HalberdsReign: 21579,
        DraculsGrasp: 21583,
        Frostwind: 21584,
        TemplarsMight: 21585,
        EschutasTemper: 21586, // also 21620?
        FirelizardsTalons: 21587,
        SandstormTrek: 21588,
        Marrowwalk: 21589,
        HeavensLight: 21590,
        ArachnidMesh: 21592,
        NosferatusCoil: 21593,
        Verdungos: 21595,
        VerdungosHeartyCord: 21595,
        CarrionWind: 21597,
        GiantSkull: 21598,
        AstreonsIronWard: 21599,
        SaracensChance: 21608,
        HighlordsWrath: 21609,
        Ravenfrost: 21610,
        Dwarfstar: 21611,
        AtmasScarab: 21612,
        Maras: 21613,
        MarasKaleidoscope: 21613,
        CrescentMoonAmulet: 21614,
        TheRisingSun: 21615,
        TheCatsEye: 21616,
        BulKathosWeddingBand: 21617,
        Metalgrid: 21619,
        Stormshield: 21621,
        BlackoakShield: 21622,
        ArkainesValor: 21624,
        TheGladiatorsBane: 21625,
        HarlequinsCrest: 21627,
        GuardianAngel: 21632,
        TheGrandfather: 21643,
        Doombringer: 21644,
        TyraelsMight: 21645,
        Lightsabre: 21646,
        TheCraniumBasher: 21647,
        DeathsWeb: 21650,
        TheAtlantean: 21654,
        CarinShard: 21658,
        Coldkill: 21286,
        ButchersCleaver: 21287,
        Islestrike: 21289,
        GuardianNaga: 21291,
        SpellSteel: 21293,
        SuicideBranch: 21297,
        ArmofKingLeoric: 21299,
        BlackhandKey: 21300,
        DarkClanCrusher: 21301,
        TheFetidSprinkler: 21304,
        HandofBlessedLight: 21305,
        Fleshrender: 21306,
        SureshrillFrost: 21307,
        Moonfall: 21308,
        BaezilsVortex: 21309,
        Earthshaker: 21310,
        TheGavelofPain: 21312,
        Bloodletter: 21313,
        ColdstealEye: 21314,
        Hexfire: 21315,
        BladeofAliBaba: 21316,
        Riftslash: 21317,
        Headstriker: 21318,
        PlagueBearer: 21319,
        //TheAtlantean: 21320,
        CrainteVomir: 21321,
        BingSzWang: 21322,
        TheVileHusk: 21323,
        Cloudcrack: 21324,
        TodesfaelleFlamme: 21325,
        Swordguard: 21326,
        Spineripper: 21327,
        HeartCarver: 21328,
        BlackbogsSharp: 21329,
        Stormspike: 21330,
        TheImpaler: 21331,
        HoneSudan: 21334,
        SpireofHonor: 21335,
        TheMeatScraper: 21336,
        BlackleachBlade: 21337,
        AthenasWrath: 21338,
        PierreTombaleCouant: 21339,
        GrimsBurningDead: 21341,
        Ribcracker: 21342,
        ChromaticIre: 21343,
        Warspear: 21344,
        SkullCollector: 21345,
        Skystrike: 21346,
        //WitchwildString: 21349,
        GoldstrikeArch: 21350,
        PusSpitter: 21352,
        VampireGaze: 21354,
        StringofEars: 21355,
        GoreRider: 21356,
        LavaGout: 21357,
        VenomGrip: 21358,
        Visceratuant: 21359,
        //GuardianAngel: 21360,
        Shaftstop: 21361,
        SkinofVipermagi: 21362,
        Blackhorn: 21363,
        ValkyrieWing: 21364,
        PeasantCrown: 21365,
        DemonMachine: 21366,
        Riphook: 21369,
        Razorswitch: 21370,
        OndalsWisdom: 21375,
        Deathbit: 21379,
        Warshrike: 21380,
        DemonLimb: 21387,
        SteelShade: 21388,
        TombReaver: 21389,
        //DeathsWeb: 21390,
        AngelsSong: 21393,
        TheRedeemer: 21394,
        Bonehew: 21398,
        Steelrend: 21399,
        AriocsNeedle: 21402,
        SoulDrainer: 21407,
        RuneMaster: 21408,
        DeathCleaver: 21409,
        ExecutionersJustice: 21410,
        Leviathan: 21412,
        WispProjector: 21417,
        Lacerator: 21419,
        MangSongsLesson: 21420,
        Viperfork: 21421,
        TheReapersToll: 21427,
        SpiritKeeper: 21428,
        Hellrack: 21429,
        AlmaNegra: 21430,
        DarkforceSpawn: 21431,
        Ghostflame: 21438,
        ShadowKiller: 21439,
        GriffonsEye: 21442,
        Thunderstroke: 21445,
        DemonsArch: 21447,
        DjinnSlayer: 21450,
        GinthersRift: 21829,

        // Runewords
        AncientsPledge: 20507,
        Armageddon: 20508,
        Authority: 20509,
        Beast: 20510,
        Beauty: 20511,
        Black: 20512,
        Blood: 20513,
        Bone: 20514,
        Bramble: 20515,
        Brand: 20516,
        BreathoftheDying: 20517,
        BrokenPromise: 20518,
        CalltoArms: 20519,
        ChainsofHonor: 20520,
        Chance: 20521,
        Chaos: 20522,
        CrescentMoon: 20523,
        Darkness: 20524,
        Daylight: 20525,
        Death: 20526,
        Deception: 20527,
        Delerium: 20528,
        Desire: 20529,
        Despair: 20530,
        Destruction: 20531,
        Doom: 20532,
        Dragon: 20533,
        Dread: 20534,
        Dream: 20535,
        Duress: 20536,
        Edge: 20537,
        Elation: 20538,
        Enigma: 20539,
        Enlightenment: 20540,
        Envy: 20541,
        Eternity: 20542,
        Exile: 20543,
        Faith: 20544,
        Famine: 20545,
        Flame: 20546,
        Fortitude: 20547,
        Fortune: 20548,
        Friendship: 20549,
        Fury: 20550,
        Gloom: 20551,
        Grief: 20553,
        HandofJustice: 20554,
        Harmony: 20555,
        HeartoftheOak: 20557,
        HolyThunder: 20560,
        Honor: 20561,
        Revenge: 20562,
        Humility: 20563,
        Hunger: 20564,
        Ice: 20565,
        Infinity: 20566,
        Innocence: 20567,
        Insight: 20568,
        Jealousy: 20569,
        Judgement: 20570,
        KingsGrace: 20571,
        Kingslayer: 20572,
        KnightsVigil: 20573,
        Knowledge: 20574,
        LastWish: 20575,
        Law: 20576,
        Lawbringer: 20577,
        Leaf: 20578,
        Lightning: 20579,
        Lionheart: 20580,
        Lore: 20581,
        Love: 20582,
        Loyalty: 20583,
        Lust: 20584,
        Madness: 20585,
        Malice: 20586,
        Melody: 20587,
        Memory: 20588,
        Mist: 20589,
        Morning: 20590,
        Mystery: 20591,
        Myth: 20592,
        Nadir: 20593,
        NaturesKingdom: 20594,
        Night: 20595,
        Oath: 20596,
        Obedience: 20597,
        Oblivion: 20598,
        Obsession: 20599,
        Passion: 20600,
        Patience: 20601,
        Patter: 20602,
        Peace: 20603,
        VoiceofReason: 20604,
        Penitence: 20605,
        Peril: 20606,
        Pestilence: 20607,
        Phoenix: 20608,
        Piety: 20609,
        PillarofFaith: 20610,
        Plague: 20611,
        Praise: 20612,
        Prayer: 20613,
        Pride: 20614,
        Principle: 20615,
        ProwessinBattle: 20616,
        Prudence: 20617,
        Punishment: 20618,
        Purity: 20619,
        Question: 20620,
        Radiance: 20621,
        Rain: 20622,
        Reason: 20623,
        Red: 20624,
        Rhyme: 20625,
        Rift: 20626,
        Sanctuary: 20627,
        Serendipity: 20628,
        Shadow: 20629,
        ShadowofDoubt: 20630,
        Silence: 20631,
        SirensSong: 20632,
        Smoke: 20633,
        Sorrow: 20634,
        Spirit: 20635,
        Splendor: 20636,
        Starlight: 20637,
        Stealth: 20638,
        Steel: 20639,
        StillWater: 20640,
        Sting: 20641,
        Stone: 20642,
        Storm: 20643,
        Strength: 20644,
        Tempest: 20645,
        Temptation: 20646,
        Terror: 20647,
        Thirst: 20648,
        Thought: 20649,
        Thunder: 20650,
        Time: 20651,
        Tradition: 20652,
        Treachery: 20653,
        Trust: 20654,
        Truth: 20655,
        UnbendingWill: 20656,
        Valor: 20657,
        Vengeance: 20658,
        Venom: 20659,
        Victory: 20660,
        Voice: 20661,
        Void: 20662,
        War: 20663,
        Water: 20664,
        Wealth: 20665,
        Whisper: 20666,
        White: 20667,
        Wind: 20668,
        WingsofHope: 20669,
        Wisdom: 20670,
        Woe: 20671,
        Wonder: 20672,
        Wrath: 20673,
        Youth: 20674,
        Zephyr: 20675,
      },
      dialog: {
        youDoNotHaveEnoughGoldForThat: 3362
      },

      text: {
        RepairCost: 3330,
        SellValue: 3331,
        IdentifyCost: 3332,
        ItemCannotBeTradedHere: 3333,
        TradeRepair: 3334,
        Buy: 3335,
        Sell: 3336,
        Heal: 3337,
        Repair: 3338,
        NextPage: 3339,
        PreviousPage: 3340,
        Transmute: 3341,
        YourGold: 3342,
        WhichItemShouldBeImbued: 3343,
        Yes: 3344,
        No: 3345,
        Gold2: 3346,
        Sell2: 3347,
        Buy2: 3358,
        Hire: 3349,
        ToStrength: 3473,
        ToDexterity: 3474,
        Defense: 3481,
        Identify: 3350,
        Repair2: 3351,
        EnhancedDefense: 3520,
        Strength: 4060,
        Dexterity: 4062,
        Vitality: 4066,
        Energy: 4069,
        DoNotMeetLevelReqForThisGame: 5162,
        CdKeyDisabled: 5199,
        CdKeyInUseBy: 5200,
        OnlyOneInstanceAtATime: 5201,
        CdKeyIntendedForAnotherProduct: 5202,
        InvalidPassword: 5207,
        AccountDoesNotExist: 5208,
        AccountIsCorrupted: 5209,
        AccountMustBeAtLeast: 5217,
        AccountCantBeMoreThan: 5218,
        PasswordMustBeAtLeast: 5219,
        PasswordCantBeMoreThan: 5220,
        LoginError: 5224,
        UsernameMustBeAtLeast: 5231,
        UsernameIncludedIllegalChars: 5232,
        UsernameIncludedDisallowedwords: 5233,
        AccountNameAlreadyExist: 5239,
        UnableToIndentifyVersion: 5245,
        UnableToCreateAccount: 5249,
        CannotCreateGamesDeadHCChar: 5304,
        Disconnected: 5347,
        BattlenetNotResponding: 5353,
        BattlenetNotResponding2: 5354,
        HcCannotPlayWithSc: 5361,
        ScCannotPlayWithHc: 5362,
        CannotPlayInHellClassic: 5363,
        CannotPlayInNightmareClassic: 5364,
        EnhancedDamage: 10038,
        ClassicCannotPlayWithXpac: 10101,
        XpacCannotPlayWithClassic: 10102,
        LoDKeyDisabled: 10913,
        LodKeyInUseBy: 10914,
        LoDKeyIntendedForAnotherProduct: 10915,
        NonLadderCannotPlayWithLadder: 10929,
        LadderCannotPlayWithNonLadder: 10930,
        YourPositionInLineIs: 11026,
        Gateway: 11049,
        Ghostly: 11084,
        Fanatic: 11085,
        Possessed: 11086,
        Berserker: 11087,
        ExpiresIn: 11133,
        CdKeyDisabledFromRealm: 11161,
        CannotPlayInHellXpac: 21793,
        CannotPlayInNightmareXpac: 21794,
      },

      areas: {
        // Act 1
        RogueEncampment: 5055,
        BloodMoor: 5054,
        ColdPlains: 5053,
        StonyField: 5052,
        DarkWood: 5051,
        BlackMarsh: 5050,
        TamoeHighland: 5049,
        DenofEvil: 5048,
        CaveLvl1: 5047,
        UndergroundPassageLvl1: 5046,
        HoleLvl1: 5045,
        PitLvl1: 5044,
        CaveLvl2: 5043,
        UndergroundPassageLvl2: 5042,
        HoleLvl2: 5041,
        PitLvl2: 5040,
        BurialGrounds: 5039,
        Crypt: 5038,
        Mausoleum: 5037,
        ForgottenTower: 5036,
        TowerCellarLvl1: 5035,
        TowerCellarLvl2: 5034,
        TowerCellarLvl3: 5033,
        TowerCellarLvl4: 5032,
        TowerCellarLvl5: 5031,
        MonasteryGate: 5030,
        OuterCloister: 5029,
        Barracks: 5038,
        JailLvl1: 5027,
        JailLvl2: 5026,
        JailLvl3: 5025,
        InnerCloister: 5024,
        Cathedral: 5023,
        CatacombsLvl1: 5022,
        CatacombsLvl2: 5021,
        CatacombsLvl3: 5020,
        CatacombsLvl4: 5019,
        Tristram: 5018,
        MooMooFarm: 788,

        // Act 2
        LutGholein: 852,
        RockyWaste: 851,
        DryHills: 850,
        FarOasis: 849,
        LostCity: 848,
        ValleyofSnakes: 847,
        CanyonofMagic: 846,
        A2SewersLvl1: 845,
        A2SewersLvl2: 844,
        A2SewersLvl3: 843,
        HaremLvl1: 842,
        HaremLvl2: 841,
        PalaceCellarLvl1: 840,
        PalaceCellarLvl2: 839,
        PalaceCellarLvl3: 838,
        StonyTombLvl1: 837,
        HallsoftheDeadLvl1: 836,
        HallsoftheDeadLvl2: 835,
        ClawViperTempleLvl1: 834,
        StonyTombLvl2: 833,
        HallsoftheDeadLvl3: 832,
        ClawViperTempleLvl2: 831,
        MaggotLairLvl1: 830,
        MaggotLairLvl2: 829,
        MaggotLairLvl3: 828,
        AncientTunnels: 827,
        TalRashasTomb1: 826,
        TalRashasTomb2: 826,
        TalRashasTomb3: 826,
        TalRashasTomb4: 826,
        TalRashasTomb5: 826,
        TalRashasTomb6: 826,
        TalRashasTomb7: 826,
        DurielsLair: 825,
        ArcaneSanctuary: 824,

        // Act 3
        KurastDocktown: 820,
        SpiderForest: 819,
        GreatMarsh: 818,
        FlayerJungle: 817,
        LowerKurast: 816,
        KurastBazaar: 815,
        UpperKurast: 814,
        KurastCauseway: 813,
        Travincal: 812,
        SpiderCave: 810,
        SpiderCavern: 811,
        SwampyPitLvl1: 809,
        SwampyPitLvl2: 808,
        FlayerDungeonLvl1: 806,
        FlayerDungeonLvl2: 805,
        SwampyPitLvl3: 807,
        FlayerDungeonLvl3: 804,
        A3SewersLvl1: 845,
        A3SewersLvl2: 844,
        RuinedTemple: 803,
        DisusedFane: 802,
        ForgottenReliquary: 801,
        ForgottenTemple: 800,
        RuinedFane: 799,
        DisusedReliquary: 798,
        DuranceofHateLvl1: 797,
        DuranceofHateLvl2: 796,
        DuranceofHateLvl3: 795,

        // Act 4
        PandemoniumFortress: 790,
        OuterSteppes: 792,
        PlainsofDespair: 793,
        CityoftheDamned: 794,
        RiverofFlame: 791,
        ChaosSanctuary: 789,

        // Act 5
        Harrogath: 22646,
        BloodyFoothills: 22647,
        FrigidHighlands: 22648,
        ArreatPlateau: 22649,
        CrystalizedPassage: 22650,
        FrozenRiver: 22651,
        GlacialTrail: 22652,
        DrifterCavern: 22653,
        FrozenTundra: 22654,
        AncientsWay: 22655,
        IcyCellar: 22656,
        ArreatSummit: 22657,
        NihlathaksTemple: 22658,
        HallsofAnguish: 22659,
        HallsofPain: 22660,
        HallsofVaught: 22662,
        Abaddon: 21865,
        PitofAcheron: 21866,
        InfernalPit: 21867,
        WorldstoneLvl1: 22663,
        WorldstoneLvl2: 22664,
        WorldstoneLvl3: 22665,
        ThroneofDestruction: 22667,
        WorldstoneChamber: 22666,

        // Ubers
        MatronsDen: 5389,
        ForgottenSands: 5389,
        FurnaceofPain: 5389,
        UberTristram: 5018,
      },
    },

    game: {
      profiletype: {
        SinglePlayer: 1,
        Battlenet: 2,
        OpenBattlenet: 3,
        TcpIpHost: 4,
        TcpIpJoin: 5
      },

      controls: {
        Disabled: 4,
      },

      gametype: {
        Classic: 0,
        Expansion: 1,
      },

      // out of game locations
      locations: {
        PreSplash: 0,
        Lobby: 1,
        WaitingInLine: 2,
        LobbyChat: 3,
        CreateGame: 4,
        JoinGame: 5,
        Ladder: 6,
        ChannelList: 7,
        MainMenu: 8,
        Login: 9,
        LoginError: 10,
        LoginUnableToConnect: 11,
        CharSelect: 12,
        RealmDown: 13,
        Disconnected: 14,
        NewCharSelected: 15,
        CharSelectPleaseWait: 16,
        LobbyLostConnection: 17,
        SplashScreen: 18,
        CdKeyInUse: 19,
        SelectDifficultySP: 20,
        MainMenuConnecting: 21,
        InvalidCdKey: 22,
        CharSelectConnecting: 23,
        ServerDown: 24,
        LobbyPleaseWait: 25,
        GameNameExists: 26,
        GatewaySelect: 27,
        GameDoesNotExist: 28,
        CharacterCreate: 29,
        OkCenteredErrorPopUp: 30,
        TermsOfUse: 31,
        CreateNewAccount: 32,
        PleaseRead: 33,
        RegisterEmail: 34,
        Credits: 35,
        Cinematics: 36,
        CharChangeRealm: 37,
        GameIsFull: 38,
        OtherMultiplayer: 39,
        TcpIp: 40,
        TcpIpEnterIp: 41,
        CharSelectNoChars: 42,
        CharSelectChangeRealm: 43,
        TcpIpUnableToConnect: 44,
      },
    },
        
    colors: {
      White: "ÿc0",
      Red: "ÿc1",
      NeonGreen: "ÿc2",
      Blue: "ÿc3",
      DarkGold: "ÿc4",
      Gray: "ÿc5",
      Black: "ÿc6",
      LightGold: "ÿc7",
      Orange: "ÿc8",
      Yellow: "ÿc9",
      DarkGreen: "ÿc:",
      Purple: "ÿc;",
      Green: "ÿc<",
      D2Bot: {
        Black: 0,
        Blue: 4,
        Green: 5,
        Gold: 6,
        DarkGold: 7,
        Orange: 8,
        Red: 9,
        Gray: 10
      }
    },

    keys: {
      Backspace: 8,
      Tab: 9,
      Enter: 13,
      Shift: 16,
      Ctrl: 17,
      Alt: 18,
      PauseBreak: 19,
      CapsLock: 20,
      Escape: 27,
      Spacebar: 32,
      PageUp: 33,
      PageDown: 34,
      End: 35,
      Home: 36,
      LeftArrow: 37,
      UpArrow: 38,
      RightArrow: 39,
      DownArrow: 40,
      Insert: 45,
      Delete: 46,
      Zero: 48,
      One: 49,
      Two: 50,
      Three: 51,
      Four: 52,
      Five: 53,
      Six: 54,
      Seven: 55,
      Eight: 56,
      Nine: 57,
      LeftWindowKey: 91,
      RightWindowKey: 92,
      SelectKey: 93,
      Numpad0: 96,
      Numpad1: 97,
      Numpad2: 98,
      Numpad3: 99,
      Numpad4: 100,
      Numpad5: 101,
      Numpad6: 102,
      Numpad7: 103,
      Numpad8: 104,
      Numpad9: 105,
      NumpadStar: 106,
      NumpadPlus: 107,
      NumpadDash: 109,
      NumpadDecimal: 110,
      NumpadSlash: 111,
      F1: 112,
      F2: 113,
      F3: 114,
      F4: 115,
      F5: 116,
      F6: 117,
      F7: 118,
      F8: 119,
      F9: 120,
      F10: 121,
      F11: 122,
      F12: 123,
      NumLock: 144,
      ScrollLock: 145,
      SemiColon: 186,
      EqualSign: 187,
      Comma: 188,
      Dash: 189,
      Period: 190,
      ForwardSlash: 191,
      GraveAccent: 192,
      OpenBracket: 219,
      BackSlash: 220,
      CloseBracket: 221,
      SingleQuote: 222,
      code: {
        Backspace: 0x08,
        Tab: 0x09,
        Clear: 0x0C,
        Enter: 0x0D,
        Shift: 0x10,
        Ctrl: 0x11,
        Alt: 0x12,
        PauseBreak: 0x13,
        CapsLock: 0x14,
        Esc: 0x1B,
        Space: 0x20,
        PageUp: 0x21,
        PageDown: 0x22,
        End: 0x23,
        Home: 0x24,
        LeftArrow: 0x25,
        UpArrow: 0x26,
        RightArrow: 0x27,
        DownArrow: 0x28,
        Select: 0x29,
        Print: 0x2A,
        PrintScreen: 0x2C,
        Insert: 0x2D,
        Delete: 0x2E,
      }
    },

    controls: {
      TextBox: 1,
      Image1: 2,
      Image2: 3,
      LabelBox: 4,
      ScrollBar: 5,
      Button: 6,
      List: 7,
      Timer: 8,
      Smack: 9,
      ProgressBar: 10,
      Popup: 11,
      AccountList: 12
    },

    packets: {
      send: {
        WalkToLocation: 0x01,
        WalkToEntity: 0x02,
        RunToLocation: 0x03,
        RunToEntity: 0x04,
        LeftSkillOnLocation: 0x05,
        LeftSkillOnEntity: 0x06,
        LeftSkillOnEntityEx: 0x07,
        LeftSkillOnLocationEx: 0x08,
        LeftSkillOnEntityEx2: 0x09,
        LeftSkillOnEntityEx3: 0x0A,
        RightSkillOnLocation: 0x0C,
        RightSkillOnEntity: 0x0D,
        RightSkillOnEntityEx: 0x0E,
        RightSkillOnLocationEx: 0x0F,
        RightSkillOnEntityEx2: 0x10,
        RightSkillOnEntityEx3: 0x11,
        SetInfernoState: 0x12,
        InteractWithEntity: 0x13,
        OverheadMessage: 0x14,
        Chat: 0x15,
        PickupItem: 0x16,
        DropItem: 0x17,
        ItemToBuffer: 0x18,
        PickupBufferItem: 0x19,
        ItemToBody: 0x1A,
        Swap2HandedItem: 0x1B,
        PickupBodyItem: 0x1C,
        SwitchBodyItem: 0x1D,
        Switch1HandWith2Hand: 0x1E,
        SwitchInventoryItem: 0x1F,
        UseItem: 0x20,
        StackItem: 0x21,
        RemoveStackItem: 0x22,
        ItemToBelt: 0x23,
        RemoveBeltItem: 0x24,
        SwitchBeltItem: 0x25,
        UseBeltItem: 0x26,
        IndentifyItem: 0x27,
        InsertSocketItem: 0x28,
        ScrollToMe: 0x29,
        ItemToCube: 0x2A,
        NPCInit: 0x2F,
        NPCCancel: 0x30,
        QuestMessage: 0x31,
        NPCBuy: 0x32,
        NPCSell: 0x33,
        NPCIndentifyItems: 0x34,
        Repair: 0x35,
        HireMerc: 0x36,
        IndentifyGamble: 0x37,
        EntityAction: 0x38,
        AddStat: 0x3A,
        AddSkill: 0x3B,
        SelectSkill: 0x3C,
        ActivateItem: 0x3E,
        CharacterPhrase: 0x3F,
        UpdateQuests: 0x40,
        Resurrect: 0x41,
        StaffInOrifice: 0x44,
        MercInteract: 0x46,
        MercMove: 0x47,
        BusyStateOff: 0x48,
        Waypoint: 0x49,
        RequestEntityUpdate: 0x4B,
        Transmorgify: 0x4C,
        PlayNPCMessage: 0x4D,
        ClickButton: 0x4F,
        DropGold: 0x50,
        BindHotkey: 0x51,
        StaminaOn: 0x53,
        StaminaOff: 0x54,
        QuestCompleted: 0x58,
        MakeEntityMove: 0x59,
        SquelchHostile: 0x5D,
        Party: 0x5E,
        UpdatePlayerPos: 0x5F,
        SwapWeapon: 0x60,
        MercItem: 0x61,
        MercRessurect: 0x62,
        LeaveGame: 0x69,
      },
      recv: {
        GameExit: 0x06,
        MapReveal: 0x07,
        MapHide: 0x08,
        ReassignPlayer: 0x15,
        SetSkill: 0x23,
        Chat: 0x26,
        UniqueEvents: 0x89,
        WeaponSwitch: 0x97,
      }
    }
  };

  // Need to be set after its loaded
  sdk.skillTabs = {
    amazon: {
      bowandcrossbow: {
        id: 0,
        skills: [
          sdk.skills.MagicArrow, sdk.skills.FireArrow, sdk.skills.MultipleShot,
          sdk.skills.ExplodingArrow, sdk.skills.IceArrow, sdk.skills.GuidedArrow,
          sdk.skills.ImmolationArrow, sdk.skills.Strafe
        ],
      },
      passiveandmagic: {
        id: 1,
        skills: [
          sdk.skills.InnerSight, sdk.skills.CriticalStrike, sdk.skills.Dodge,
          sdk.skills.SlowMissiles, sdk.skills.Avoid, sdk.skills.Penetrate,
          sdk.skills.Dopplezon, sdk.skills.Evade, sdk.skills.Valkyrie, sdk.skills.Pierce
        ],
      },
      javelinandspear: {
        id: 2,
        skills: [
          sdk.skills.Jab, sdk.skills.PowerStrike, sdk.skills.PoisonJavelin,
          sdk.skills.Impale, sdk.skills.LightningBolt, sdk.skills.ChargedStrike,
          sdk.skills.PlagueJavelin, sdk.skills.Fend, sdk.skills.LightningStrike, sdk.skills.LightningFury
        ],
      },
    },
    sorc: {
      fire: {
        id: 8,
        skills: [
          sdk.skills.FireBolt, sdk.skills.Warmth, sdk.skills.Inferno,
          sdk.skills.Blaze, sdk.skills.FireBall, sdk.skills.FireWall,
          sdk.skills.Enchant, sdk.skills.Meteor, sdk.skills.FireMastery, sdk.skills.Hydra
        ],
      },
      lightning: {
        id: 9,
        skills: [
          sdk.skills.ChargedBolt, sdk.skills.StaticField, sdk.skills.Telekinesis,
          sdk.skills.Nova, sdk.skills.Lightning, sdk.skills.ChainLightning,
          sdk.skills.Teleport, sdk.skills.ThunderStorm, sdk.skills.EnergyShield,
          sdk.skills.LightningMastery
        ],
      },
      cold: {
        id: 10,
        skills: [
          sdk.skills.IceBolt, sdk.skills.FrozenArmor, sdk.skills.FrostNova,
          sdk.skills.IceBlast, sdk.skills.ShiverArmor, sdk.skills.GlacialSpike,
          sdk.skills.Blizzard, sdk.skills.ChillingArmor,
          sdk.skills.FrozenOrb, sdk.skills.ColdMastery
        ]
      }
    },
    necro: {
      curse: {
        id: 16,
        skills: [
          sdk.skills.AmplifyDamage, sdk.skills.DimVision, sdk.skills.Weaken,
          sdk.skills.IronMaiden, sdk.skills.Terror, sdk.skills.Confuse,
          sdk.skills.LifeTap, sdk.skills.Attract,
          sdk.skills.Decrepify, sdk.skills.LowerResist
        ],
      },
      poisonandbone: {
        id: 17,
        skills: [
          sdk.skills.Teeth, sdk.skills.BoneArmor, sdk.skills.PoisonDagger,
          sdk.skills.CorpseExplosion, sdk.skills.BoneWall, sdk.skills.PoisonExplosion,
          sdk.skills.BoneSpear, sdk.skills.BonePrison,
          sdk.skills.PoisonNova, sdk.skills.BoneSpirit
        ],
      },
      summoning: {
        id: 18,
        skills: [
          sdk.skills.SkeletonMastery, sdk.skills.RaiseSkeleton, sdk.skills.ClayGolem,
          sdk.skills.GolemMastery, sdk.skills.RaiseSkeletalMage, sdk.skills.BloodGolem,
          sdk.skills.SummonResist, sdk.skills.IronGolem,
          sdk.skills.FireGolem, sdk.skills.Revive
        ]
      }
    },
    paladin: {
      combat: {
        id: 24,
        skills: [
          sdk.skills.Sacrifice, sdk.skills.Smite, sdk.skills.HolyBolt,
          sdk.skills.Zeal, sdk.skills.Charge, sdk.skills.Vengeance,
          sdk.skills.BlessedHammer, sdk.skills.Conversion,
          sdk.skills.HolyShield, sdk.skills.FistoftheHeavens
        ],
      },
      offensiveaura: {
        id: 25,
        skills: [
          sdk.skills.Might, sdk.skills.HolyFire, sdk.skills.Thorns,
          sdk.skills.BlessedAim, sdk.skills.Concentration, sdk.skills.HolyFreeze,
          sdk.skills.HolyShock, sdk.skills.Sanctuary,
          sdk.skills.Fanaticism, sdk.skills.Conviction
        ],
      },
      defensiveaura: {
        id: 26,
        skills: [
          sdk.skills.Prayer, sdk.skills.ResistFire, sdk.skills.Defiance,
          sdk.skills.ResistCold, sdk.skills.Cleansing, sdk.skills.ResistLightning,
          sdk.skills.Vigor, sdk.skills.Meditation,
          sdk.skills.Redemption, sdk.skills.Salvation
        ],
      }
    },
    barb: {
      combat: {
        id: 32,
        skills: [
          sdk.skills.Bash, sdk.skills.Leap, sdk.skills.DoubleSwing,
          sdk.skills.Stun, sdk.skills.DoubleThrow, sdk.skills.LeapAttack,
          sdk.skills.Concentrate, sdk.skills.Frenzy,
          sdk.skills.Whirlwind, sdk.skills.Berserk
        ],
      },
      masteries: {
        id: 33,
        skills: [
          sdk.skills.SwordMastery, sdk.skills.AxeMastery, sdk.skills.MaceMastery, sdk.skills.PoleArmMastery,
          sdk.skills.ThrowingMastery, sdk.skills.SpearMastery, sdk.skills.IncreasedStamina,
          sdk.skills.IronSkin, sdk.skills.IncreasedSpeed, sdk.skills.NaturalResistance
        ],
      },
      warcries: {
        id: 34,
        skills: [
          sdk.skills.BattleOrders, sdk.skills.BattleCommand, sdk.skills.WarCry, sdk.skills.BattleCry,
          sdk.skills.GrimWard, sdk.skills.FindItem, sdk.skills.Shout,
          sdk.skills.Taunt, sdk.skills.Howl, sdk.skills.FindPotion
        ],
      }
    },
    druid: {
      summoning: {
        id: 40,
        skills: [
          sdk.skills.Raven, sdk.skills.PoisonCreeper, sdk.skills.OakSage,
          sdk.skills.SummonSpiritWolf, sdk.skills.CarrionVine, sdk.skills.HeartofWolverine,
          sdk.skills.SummonDireWolf, sdk.skills.SolarCreeper, sdk.skills.SpiritofBarbs, sdk.skills.SummonGrizzly
        ],
      },
      shapeshifting: {
        id: 41,
        skills: [
          sdk.skills.Werewolf, sdk.skills.Lycanthropy, sdk.skills.Werebear, sdk.skills.FeralRage,
          sdk.skills.Maul, sdk.skills.Rabies, sdk.skills.FireClaws,
          sdk.skills.Hunger, sdk.skills.ShockWave, sdk.skills.Fury
        ],
      },
      elemental: {
        id: 42,
        skills: [
          sdk.skills.Firestorm, sdk.skills.MoltenBoulder, sdk.skills.ArcticBlast,
          sdk.skills.Fissure, sdk.skills.CycloneArmor, sdk.skills.Twister,
          sdk.skills.Volcano, sdk.skills.Tornado, sdk.skills.Armageddon, sdk.skills.Hurricane
        ]
      }
    },
    assassin: {
      traps: {
        id: 48,
        skills: [
          sdk.skills.FireBlast, sdk.skills.ShockWeb, sdk.skills.BladeSentinel,
          sdk.skills.ChargedBoltSentry, sdk.skills.WakeofFire, sdk.skills.BladeFury,
          sdk.skills.LightningSentry, sdk.skills.WakeofInferno, sdk.skills.DeathSentry, sdk.skills.BladeShield
        ],
      },
      shadowdisciplines: {
        id: 49,
        skills: [
          sdk.skills.ClawMastery, sdk.skills.PsychicHammer, sdk.skills.BurstofSpeed,
          sdk.skills.WeaponBlock, sdk.skills.CloakofShadows, sdk.skills.Fade,
          sdk.skills.ShadowWarrior, sdk.skills.MindBlast, sdk.skills.Venom, sdk.skills.ShadowMaster
        ],
      },
      martialarts: {
        id: 50,
        skills: [
          sdk.skills.TigerStrike, sdk.skills.DragonTalon, sdk.skills.FistsofFire,
          sdk.skills.DragonClaw, sdk.skills.CobraStrike, sdk.skills.ClawsofThunder,
          sdk.skills.DragonTail, sdk.skills.BladesofIce, sdk.skills.DragonFlight, sdk.skills.PhoenixStrike
        ],
      }
    },
  };

  return sdk;
}));
